"""Experimental features.

This module includes experimental features such as those which
rely on the numba jitclass decorator (which itself is experimental).
"""

import logging

import numba
import numpy
import pandas

from fast_tx import buffer_tools


def apply_chain_jc(dataframe: pandas.DataFrame,
                   jc_list):
    """Apply given transform chain (of jitted classes) to dataframe.

    :param dataframe:  DataFrame to modify.

    :param jc_list:   List numba jitted classes with calc_row methods
                      to apply to dataframe.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    PURPOSE:  Apply list of jitted classes in `jc_list` to the given
              `dataframe`.
    """
    input_cols, output_cols, maincols = buffer_tools.prepare_cols(
        dataframe, jc_list)
    index, input_buffer, output_buffer, mainbuffer = buffer_tools.make_buffers(
        dataframe, maincols, input_cols, output_cols)

    # If original dataframe has mixed types, mainbuffer may not be float.
    # We must force float for numba, but warn the user about that so they
    # can maybe try to fix type themselves.

    if mainbuffer.dtype != 'float64':  # pragma: no cover
        logging.warning('Buffer dataframe has type %s; will force float64',
                        mainbuffer.dtype)
        mainbuffer = mainbuffer.astype(numpy.float64)

    _apply_inner_loop_jc(jc_list, index.astype(numpy.float64),
                         mainbuffer, input_buffer, output_buffer,
                         input_cols, output_cols)
    dataframe[list(maincols)] = mainbuffer


def _apply_inner_loop_jc(
        # pylint: disable=too-many-arguments,too-many-positional-arguments
        jc_list, index, mainbuffer, input_buffer, output_buffer,
        input_cols, output_cols):
    for r_num, row in enumerate(mainbuffer):
        t_num = 0
        for tx_func in numba.literal_unroll(jc_list):
            i = 0
            while input_cols[t_num][i] >= 0:
                input_buffer[i] = row[input_cols[t_num, i]]
                i += 1
            tx_func.calc_row(index[r_num], input_buffer[:i],
                             output_buffer[:-output_cols[t_num, -1]])
            i = 0
            while output_cols[t_num][i] >= 0:
                row[output_cols[t_num][i]] = output_buffer[i]
                i += 1
            t_num += 1



"""Small utility functions.
"""

import datetime

import numpy
import pandas


def date_to_utctimestamp(my_date) -> float:
    """Convert input date to float utc timestamp

    Useful to get things into float form for numba functions.
    """
    if pandas.isnull(my_date):
        return numpy.nan
    return datetime.datetime(my_date.year, my_date.month, my_date.day,
                             tzinfo=datetime.timezone.utc).timestamp()



"""Simple profiling tests for Exponentially Weighted Moving Average.
"""


import cProfile
import io
import os
import pstats
from pstats import SortKey
import timeit

import numba
from numba.types import UniTuple, Tuple
import numpy
import pandas
import pytest



from fast_tx.core import (FastTx, SimpleEWA, apply_chain, NDArray)


class CleanEWA(FastTx):#FIXME: need docs

    def __init__(self, alphas, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.alphas = numpy.asarray(alphas)

    def startup(self, dataframe):
        state = numpy.zeros(2*len(self._output_col_names))
        state[:len(self._output_col_names)] = self.alphas
        return state

    @staticmethod
    def calc_row_py(index_value, state, inputs, outputs):
        alphas = state[:len(outputs)]
        prev_out = state[len(outputs):]
        outputs[:] = alphas * inputs + (1-alphas) * prev_out
        state[len(outputs):] = outputs[:]


class FullEWA(FastTx):#FIXME: need docs

    def __init__(self, alphas, *args, **kwargs):
        super().__init__(*([args[0]+args[1]] + list(args[1:])), **kwargs)
        self.alphas = numpy.asarray(alphas)

    def startup(self, dataframe):
        state = numpy.zeros(2*len(self._output_col_names))
        state[:len(self._output_col_names)] = self.alphas
        return state

    @staticmethod
    def calc_row_py(index_value, state, inputs, outputs):
        l_o = len(outputs)
        alphas = state[:l_o]
        prev_out = state[l_o:]
        outputs[:] = alphas * inputs[:l_o] + (1-alphas) * prev_out

    @staticmethod
    def post_process_state_py(index_value, state, inputs):
        l_o = len(inputs)//2
        state[l_o:] = inputs[l_o:]


class ClosureEWA(FastTx):#FIXME: need docs

    def __init__(self, alphas, *args, **kwargs):
        super().__init__(*([args[0]+args[1]] + list(args[1:])), **kwargs)
        self.alphas = numpy.asarray(alphas)

        alphas = self.alphas
        l_o = len(alphas)
        my_state = numpy.zeros(l_o)
        #print(f'making {l_o=}')
        foo = 5.0
        @numba.njit(numba.void(numba.float64, numba.float64[:],
                               numba.float64[:], numba.float64[:]),
                    cache=True)
        def calc_row_py(index_value, state, inputs, outputs):
            prev_out = state[l_o:]
            outputs[:] = alphas * inputs[:l_o] + (1-alphas) * prev_out
        self.calc_row = calc_row_py

    def startup(self, dataframe):
        state = numpy.zeros(2*len(self._output_col_names))
        state[:len(self._output_col_names)] = self.alphas
        return state

    @staticmethod
    def post_process_state_py(index_value, state, inputs):
        l_o = len(inputs)//2
        state[l_o:] = inputs[l_o:]


class CrazyEWA(FastTx):#FIXME: need docs

    def __init__(self, alphas, *args, **kwargs):
        super().__init__(*([args[0]+args[1]] + list(args[1:])), **kwargs)
        self.alphas = numpy.asarray(alphas)

        alphas = self.alphas
        l_o = len(alphas)
        my_state = numpy.zeros(l_o)
        #print(f'making {l_o=}')


        @numba.njit(Tuple((
            numba.float64, numba.float64[:], numba.float64[:]))(
                numba.float64[:]), cache=True)
        def unpack(state):
            l_o = len(state)//2
            alphas = state[:l_o]        
            prev_out = state[l_o:]
            return l_o, alphas, prev_out


        @numba.njit(numba.void(numba.float64, numba.float64[:],
                               numba.float64[:], numba.float64[:]),
                    cache=False)
        def calc_row_py(index_value, state, inputs, outputs):
            l_o, alphas, prev_out = unpack(state)
            outputs[:] = alphas * inputs[:l_o] + (1-alphas) * prev_out
        self.calc_row = calc_row_py



    def startup(self, dataframe):
        state = numpy.zeros(2*len(self._output_col_names))
        state[:len(self._output_col_names)] = self.alphas
        return state

    @staticmethod
    def post_process_state_py(index_value, state, inputs):
        l_o = len(inputs)//2
        state[l_o:] = inputs[l_o:]


class DynamicClosureEWA(FastTx):#FIXME: need docs

    def __init__(self, alphas, input_col_names, output_col_names, **kwargs):
        assert len(input_col_names) == len(output_col_names)
        assert len(alphas) == len(input_col_names)
        self.alphas = numpy.asarray(alphas)
        self.patch_calc_row()
        super().__init__(input_col_names, output_col_names, **kwargs)

    def startup(self, dataframe):
        # We store alphas at end of state array.
        # We will use an `unpack` closure to extract in modify_calc_row
        state = numpy.zeros(2*len(self._output_col_names))
        state[:len(self._output_col_names)] = self.alphas
        return state
        
    def patch_calc_row(self):
        l_o = len(self.alphas)  # can define variable for closure

        @numba.njit(Tuple((
            numba.float64, numba.float64[:], numba.float64[:]))(
                numba.float64[:]), cache=True)
        def unpack(state):
            alphas = state[:l_o]  # l_o is closure variable
            prev_out = state[l_o:]
            return l_o, alphas, prev_out

        @numba.njit(numba.void(numba.float64, numba.float64[:],
                               numba.float64[:], numba.float64[:]),
                    cache=False)
        def calc_row_py(index_value, state, inputs, outputs):
            l_o, alphas, prev_out = unpack(state)
            outputs[:] = alphas * inputs[:l_o] + (1-alphas) * prev_out
            prev_out[:] = outputs[:]  # updates state for next row calc
        self.calc_row = calc_row_py

class HackDynamicClosureEWA(FastTx):#FIXME: need docs

    def __init__(self, alphas, input_col_names, output_col_names, **kwargs):
        assert len(input_col_names) == len(output_col_names)
        assert len(alphas) == len(input_col_names)
        self.alphas = numpy.asarray(alphas)
        self.make_tx()
        super().__init__(input_col_names, output_col_names, **kwargs)

    def startup(self, dataframe):
        # We store alphas at end of state array.
        # We will use an `unpack` closure to extract in modify_calc_row
        state = numpy.zeros(2*len(self._output_col_names))
        state[:len(self._output_col_names)] = self.alphas
        return state
        
    def make_tx(self):
        @numba.njit(Tuple((numba.float64, numba.float64[:],
                           numba.float64[:]))(numba.float64[:]))
        def unpack(state):            
            l_o = len(state)//2
            alphas = state[:l_o]
            prev_out = state[l_o:]
            return l_o, alphas, prev_out

        @numba.njit
        def calc_row_py(index_value, state, inputs, outputs):  # can reference l_o above
            l_o, alphas, prev = unpack(state)
            outputs[:] = alphas * inputs + (1-alphas) * prev
            state[l_o:] = outputs[:]

        self.calc_row = calc_row_py
    

EWM_ENGINE_TO_CLS = {
    'fast_tx': SimpleEWA,
    'fast_tx_clean': CleanEWA,
    'fast_tx_full': FullEWA,
    'fast_tx_closure': ClosureEWA,
    'fast_tx_dynamic_closure': DynamicClosureEWA,
    'fast_tx_crazy': CrazyEWA,
    'fast_tx_hack': HackDynamicClosureEWA,
    }

def check_ewm(df, engine, incol='data', outcol='result', alpha=0.1,
              tx_list=None, verify=False):

    if tx_list is not None:
        apply_chain(df, tx_list)
    elif engine in EWM_ENGINE_TO_CLS:
        tx_list = [EWM_ENGINE_TO_CLS[engine]([alpha], [incol], [outcol])]
        apply_chain(df, tx_list)
    elif engine == 'apply_closure':
        def make_func(my_alpha, my_incol):
            prev = [0.0]  # need mutable container or nonlocal
            # performance is same for nonlocal or list or numpy array
            def my_func(row):
                prev[0] = row[my_incol]*my_alpha + (1-my_alpha)*prev[0]
                return prev[0]
            return my_func
        df[outcol] = df.apply(make_func(alpha, incol), axis=1)
    elif engine == 'apply_cls':
        class MyEWM:
            def __init__(self):
                self.prev = 0
            def __call__(self, row):
                self.prev = row[incol]*alpha + (1-alpha)*self.prev
                return self.prev
        my_thing = MyEWM()
        df[outcol] = df.apply(my_thing, axis=1)
    elif engine in ('numba', 'cython'):
        df[outcol] = pandas.concat([pandas.DataFrame(
            {incol: 0, outcol: numpy.nan}, index=[-1]), df])[incol].ewm(
                alpha=alpha, adjust=0).mean(engine=engine).iloc[1:]
    else:
        raise ValueError(f'unknown {engine=}')
    if verify:
        check_accuracy(df, incol, outcol, alpha)
    return tx_list
    

def check_accuracy(my_frame, incol='data', outcol='result', alpha=0.1):
    check_frame = my_frame.copy()
    check_frame[outcol] = numpy.nan
    check_frame = pandas.concat([pandas.DataFrame(
        {incol: 0, outcol: numpy.nan}, index=[-1]), check_frame])
    check_frame[outcol] = check_frame[incol].ewm(alpha=alpha, adjust=0
                                                 ).mean()
    assert all((check_frame.iloc[1:] == my_frame).all())


def prep_for_check_ewm(preprocess_compile, engine=None, size=136,
                       verify=False):
    engine = engine or 'fast_tx'
    size = int(os.environ.get('FAST_TX_TEST_SIZE', size))
    state = numpy.random.RandomState(123)  # pylint: disable=no-member
    data = 100 * numpy.cumprod(state.lognormal(0, .01, size))
    frame = pandas.DataFrame({'data': data})
    frame['result'] = numpy.nan
    
    if preprocess_compile:
        tx_list = check_ewm(frame.iloc[:10].copy(), engine)

    return {'df': frame, 'engine': engine, 'tx_list': tx_list,
            'verify': verify}

def do_profile(preprocess_compile=False, trials=1, engine=None):  # pragma: no cover
    """Do profiling for test_price_div_big function.

    :param preprocess_compile=False:  If True, we do a small run
                                      outside the profiler to force
                                      numba to JIT compile and read
                                      from the cache first.
                                      Supposedly using `cache=True` in
                                      JIT compiling will cache things,
                                      but even with that, there still
                                      seems to be an initial startup
                                      hit. See the following thread on
                                      https://numba.discourse.group/t/:

      understanding-initial-execution-delay-in-numba-cached-functions/2783/9

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    PURPOSE:  Do profiling.

    """
    kwargs = prep_for_check_ewm(preprocess_compile, engine=engine)
    tx_list = kwargs['tx_list']
    print(f'Doing profiling with engine: ' + kwargs['engine'])
    pr = cProfile.Profile()
    pr.enable()
    for _ in range(trials):
        my_frame = kwargs['df'].copy()
        tx_list = check_ewm(my_frame, kwargs['engine'], tx_list=tx_list)
    pr.disable()
    s = io.StringIO()
    sortby = SortKey.CUMULATIVE
    ps = pstats.Stats(pr, stream=s).sort_stats(sortby)
    my_func = ps.get_stats_profile().func_profiles['check_ewm']
    check_accuracy(my_frame)
    print(my_func)
    if os.environ.get('FAST_TX_SHOW_FULL_PROFILE', '0').strip() not in [
            '1', 'yes', 'YES', 'Yes', 'true', 'True']:
        return
    ps.print_stats()
    print(s.getvalue())

    
def do_timeit(preprocess_compile=True, engine=None, number=None, **kwargs):
    kwargs = prep_for_check_ewm(preprocess_compile, engine=engine, **kwargs)
    my_timer = timeit.Timer('check_ewm(**kwargs)', globals={
        'kwargs': kwargs, 'check_ewm': check_ewm})
    if number is None:
        number, total = my_timer.autorange()
    else:
        total = my_timer.timeit(number=number)
    avg = total/float(number)
    result = (f'{avg=:.4f} seconds in {number} trials for engine ' + kwargs[
        'engine'])
    return result


def main(engine=None, doprint=print, mode=None, **kwargs):
    result = []
    mode = mode or os.environ.get('FAST_TX_PROFILE_MODE', 'timeit')
    engine = engine or os.environ.get('FAST_TX_ENGINE', 'fast_tx')
    if ',' in engine:
        result = sum([main(engine=e.strip(), doprint=doprint, mode=mode,
                           **kwargs) for e in engine.split(',')],[])
    elif engine == '__all__':
        engine = ','.join(list(EWM_ENGINE_TO_CLS) + [
            'cython', 'apply_cls', 'apply_closure'])
        return main(engine=engine, doprint=doprint, mode=mode, **kwargs)
    elif mode == 'timeit':
        result = [do_timeit(engine=engine, **kwargs)]
        doprint(result[0])
    elif mode == 'profile':
        result = [do_profile(
            preprocess_compile=True, trials=int(os.environ.get(
                'FAST_TX_NUM_TRIALS', '20')), engine=engine)]
        doprint(result[0])
    else:
        raise ValueError(f'Bad {mode=}')
    return result


def test_all():
    """Test everything in this file (mainly for coverage).

    This function will test everything in the file (mainly to ensure
    code coverage).
    """
    with pytest.raises(ValueError) as exc_info:
        main(mode='broken')        
    assert exc_info.type == ValueError
    with pytest.raises(ValueError) as exc_info:
        main(engine='broken')
    assert exc_info.type == ValueError
    main(engine='__all__', size=100, verify=False, number=None)    
    main(engine='__all__', size=10, verify=True, number=1)
    main(engine='fast_tx', size=100, mode='profile', verify=True)


if __name__ == '__main__':  # pragma: no cover
    main()



"""Some simple tests for the fast_tx package.

Note: If you have problems or want to debug, you should set NUMBA_DISABLE_JIT=1
      as an environment variable and then run tests. That will disable the
      NUMBA jit compliation so you can use pdb and get better error info.
"""

import cProfile
import io
import pstats
from pstats import SortKey

import numba
import numpy
import pandas

from fast_tx.core import (FastTx, Lag, SafeSum, apply_chain)


class TargetPosition(FastTx):
    """Compute a target position for stock.

    Target is previous holding plus use opening cash / price to buy more.
    """

    def __init__(self, input_cols=('open_cash', 'price', 'prev_shares'),
                 output_cols=('shares',)):
        super().__init__(input_cols, output_cols)

    @staticmethod
    @numba.njit(numba.void(numba.float64, numba.float64[:],
                           numba.float64[:], numba.float64[:]), cache=True)
    def calc_row(index_value, state, inputs, outputs):
        outputs[0] = round(numpy.floor(inputs[0]/inputs[1]), 2)
        if not numpy.isnan(inputs[2]):
            outputs[0] += inputs[2]


class TotalCash(FastTx):
    """Compute total cash.

    Cash is cash from the open plus previous dividends (times previous shares)
    plus cost of any shares we are buying now.
    """

    def __init__(self, input_cols=(
            'open_cash', 'prev_div', 'price', 'prev_shares', 'shares'),
                 output_cols=('close_cash',)):
        super().__init__(input_cols, output_cols)

    @staticmethod
    @numba.njit(numba.void(numba.float64, numba.float64[:],
                           numba.float64[:], numba.float64[:]), cache=True)
    def calc_row(index_value, state, inputs, outputs):
        outputs[0] = (inputs[0]              # cash we had at the open
                      + inputs[1]*inputs[3]  # dividends received
                      ) - inputs[2] * (inputs[4]-inputs[3])  # shares bought


def slow_price_div(frame):
    """Compute shares to own in a safe (but slow) way for the given frame.

    This is intended as a helper function to check the fast calculation
    using transforms.
    """
    fcopy = frame.copy()
    for i in range(1, len(fcopy.index)):
        fcopy.loc[fcopy.index[i], 'open_cash'] = (
            fcopy.iloc[i-1]['close_cash'] + fcopy.iloc[i-1]['deposit'])
        fcopy.loc[fcopy.index[i], 'shares'] = (
            round(numpy.floor(
                fcopy.iloc[i]['open_cash']/fcopy.iloc[i]['price']), 2)
            + fcopy.iloc[i-1]['shares'])
        fcopy.loc[fcopy.index[i], 'close_cash'] = fcopy.iloc[
            i]['open_cash'] - (
                fcopy.iloc[i]['shares'] - fcopy.iloc[i-1]['shares']
                ) * fcopy.iloc[i]['price'] + (
                    fcopy.iloc[i-1]['div'] * fcopy.iloc[i-1]['shares'])
    return fcopy


def check_price_div(frame, check_diff=True):
    """Helper to check price/dividend math as a test case.

    :param frame:    Input DataFrame with 'div' and 'price' columns.

    :param check_diff=True:   Whether to validate result with `slow_price_div`.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    PURPOSE:   Run some transforms on `frame` to verify that the transforms
               correctly to some price/dividind and buying math.

    """
    frame[['shares', 'open_cash', 'close_cash', 'deposit']] = 0.0
    frame.loc[frame.index[0], 'deposit'] = 100.0
    fcopy = None
    if check_diff:
        fcopy = slow_price_div(frame)

    frame[['prev_shares']] = 0
    apply_chain(frame, [
        Lag(1, ['shares', 'div', 'close_cash', 'deposit'], [
            'prev_shares', 'prev_div', 'prev_close_cash', 'prev_deposit']),
        SafeSum(['prev_close_cash', 'prev_deposit'], ['open_cash']),
        TargetPosition(), TotalCash()])

    if check_diff and fcopy is not None:
        diff = fcopy.sub(frame[list(fcopy)])
        diff = diff.loc[diff.abs().sum(axis=1) > 1e-6]
        assert diff.empty


def test_price_div_small():
    """Test small example we have verified by hand.
    """
    frame = pandas.DataFrame([
        [.5, 4],
        [0,  5],
        [.7, 6.8],
        [0, 7.5],
        [.95, 8.25]], columns=['div', 'price'])
    check_price_div(frame)


def test_price_div_big(size=int(1e2)):
    """Run a larger version of the price/div test.

    The main purpose of this test is that you can change the `size`
    parameter and run with the environment variable
    `NUMBA_DISABLE_JIT=1` (and probably use `-k test_price_div_big` to
    just test this function) and without to compare the speed of using
    numba and not using numba. This example is fairly trivial, so you
    will need large sizes (e.g., on the order of `size=int(1e6)`) to
    really see a difference.

    Note that we automatically set check_diff to False for large sizes
    so that manually checking correctness does not confuse the timing.
    """
    check_diff = size < 10000  # do not check diff for big sizes
    state = numpy.random.RandomState(123)  # pylint: disable=no-member
    prices = 100 * numpy.cumprod(state.lognormal(0, .01, size))
    divs = [(1 if i else 0) for i in (state.uniform(0, 1, size) > .9)]
    frame = pandas.DataFrame({'price': prices, 'div': divs})
    check_price_div(frame, check_diff=check_diff)


def do_profile(preprocess_compile=False, trials=1):  # pragma: no cover
    """Do profiling for test_price_div_big function.

    :param preprocess_compile=False:  If True, we do a small run
                                      outside the profiler to force
                                      numba to JIT compile and read
                                      from the cache first.
                                      Supposedly using `cache=True` in
                                      JIT compiling will cache things,
                                      but even with that, there still
                                      seems to be an initial startup
                                      hit. See the following thread on
                                      https://numba.discourse.group/t/:

      understanding-initial-execution-delay-in-numba-cached-functions/2783/9

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    PURPOSE:  Do profiling.

    """
    if preprocess_compile:
        test_price_div_big(10)

    print('Doing profiling')
    pr = cProfile.Profile()
    pr.enable()
    for _ in range(trials):
        test_price_div_big(30000)
    pr.disable()
    s = io.StringIO()
    sortby = SortKey.CUMULATIVE
    ps = pstats.Stats(pr, stream=s).sort_stats(sortby)
    ps.print_stats()
    print(s.getvalue())


if __name__ == '__main__':  # pragma: no cover
    do_profile(preprocess_compile=True, trials=20)




import typing

import numba
from numba import float64, int64, typeof
from numba.experimental import jitclass
from numba.types import ListType, unicode_type
from numba.typed import List as NumbaList
import numpy
import pandas

from fast_tx.core import FastTx
from fast_tx.experimental import apply_chain_jc


@jitclass([
    ('params', float64[:]),               
    ('_input_col_names', ListType(unicode_type)),
    ('_output_col_names', ListType(unicode_type)),
    ('state', float64[:]),
    ])
class EWM:

    def __init__(self, params, input_col_names, output_col_names):
        self.params = params
        self._input_col_names = input_col_names
        self._output_col_names = output_col_names
        self.state = numpy.zeros(len(input_col_names), dtype=numpy.float64)

    def input_col_names(self):
        return self._input_col_names

    def output_col_names(self):
        return self._output_col_names    

    def calc_row(self, index_value,
                 inputs: float64[:], outputs: float64[:]):
        self.state[:] = (1-self.params)*self.state[:] + self.params*inputs
        outputs[:] = self.state[:]


@jitclass([
    ('threshold', int64[:]),               
    ('_input_col_names', ListType(unicode_type)),
    ('_output_col_names', ListType(unicode_type)),
    ('count', int64[:]),
    ])
class OnlyFirstN:

    def __init__(self, threshold, input_col_names, output_col_names):
        self.threshold = threshold
        self._input_col_names = input_col_names
        self._output_col_names = output_col_names
        self.count = numpy.zeros(len(input_col_names), dtype=numpy.int64)

    def input_col_names(self):
        return self._input_col_names

    def output_col_names(self):
        return self._output_col_names    

    def calc_row(self, index_value,
                 inputs: float64[:], outputs: float64[:]):
        for i in range(len(inputs)):
            if self.count[i] >= self.threshold[i]:
                outputs[i] = 0  # already have N non-null so zero output
            elif numpy.isnan(inputs[i]):  # found a nan so just use 0
                outputs[i] = 0
            else:  # founda non-nan so pass to output and incr count
                outputs[i] = inputs[i]
                self.count[i] += 1


def test_jc_EWM():
    df = pandas.DataFrame({
        'a': [1.1, 2.2, 3.3, 4.4, 5.5],
        'b': [10.0, 20.5, 30.1, 40.8, 50.2],
        'c': [0.12, 0.34, 0.56, 0.78, 0.90]})
    golden = df.copy()
    zp_df = pandas.concat([pandas.DataFrame({'a':0,'b':0,'c':0},index=[-1]),
                           df])  # zero padded for use in verifying EWM
    params = [.1, .2, .3]
    i_cols = NumbaList.empty_list(unicode_type)
    for item in ('a', 'b', 'c'):
        i_cols.append(item)
    o_cols = NumbaList.empty_list(unicode_type)        
    for item in i_cols:
        o_cols.append('out_' + item)
    jc_list = (  # Needs to be a tuple for numba literal_unroll
        EWM(numpy.array(params),i_cols, o_cols),
    )
    apply_chain_jc(df, jc_list)
    for i, name in enumerate(zp_df):
        out_name = 'out_' + name
        golden[out_name] = zp_df[name].ewm(
            alpha=params[i], adjust=0).mean().iloc[1:]
        assert (golden[out_name] == df[out_name]).all()


def numba_s_list(inputs: typing.Sequence[str]):
    """Helper function to create typed list of strings that numba wants.

Numba is cranky about reflected lists and so we need to construct a list
in a way where the types are very explicit. This function takes in a
sequence of strings and produces a numba typed list of strings.    
    """
    result = NumbaList.empty_list(unicode_type)
    for item in inputs:
        result.append(item)
    return result


def test_jc_OnlyFirstN():
    df = pandas.DataFrame({
        'a': [1.1, 2.2, 3.3, 4.4, 5.5],
        'b': [10.0, 20.5, 30.1, 40.8, 50.2],
        'c': [numpy.nan, 0.34, 0.56, 0.78, 0.90]})
    golden = df.copy()
    golden[['out_' + n for n in df]] = 0.0
    golden.loc[:1, 'out_a'] = df['a'].loc[:1]
    golden.loc[:2, 'out_b'] = df['b'].loc[:2]
    golden['out_c'] = df['c']
    golden.loc[0, 'out_c'] = 0.0
    i_cols = numba_s_list(['a', 'b', 'c'])
    o_cols = numba_s_list(['out_' + i for i in i_cols])
    tx = OnlyFirstN(numpy.array([2, 3, 4]), i_cols, o_cols)
    apply_chain_jc(df, [tx])
    assert (df != golden).sum().sum() == 1  # only NaN gives a non-match
    

def test_jc_EWM_OnlyFirstN():
    """Test jitted class with two different kinds of transforms.

This function tests applying jitted classes to a DataFrame with
two different kinds of jitted class with different types of parameters
and states. This verifies that:

    1. We can use heterogeneous classes (same call signature but
       different internal state and calculation).
    2. We can apply a sequence of classes to each row.

    """
    df = pandas.DataFrame({
        'a': [1.1, 2.2, 3.3, 4.4, 5.5],
        'b': [10.0, 20.5, 30.1, 40.8, 50.2],
        'c': [0.12, 0.34, 0.56, 0.78, 0.90]})
    golden = df.copy()
    zp_df = pandas.concat([pandas.DataFrame({'a':0,'b':0,'c':0},index=[-1]),
                           df])  # zero padded for use in verifying EWM
    params = [.1, .2, .3]
    i_cols = numba_s_list(['a', 'b', 'c'])
    o_cols = numba_s_list(['out_' + i for i in i_cols])
    jc_list = (  # Needs to be a tuple for numba literal_unroll
        EWM(numpy.array(params), i_cols, o_cols),
        OnlyFirstN(numpy.array([2, 3, 4]), o_cols, o_cols),
    )
    apply_chain_jc(df, jc_list)
    for i, name in enumerate(zp_df):
        out_name = 'out_' + name
        golden[out_name] = zp_df[name].ewm(
            alpha=params[i], adjust=0).mean().iloc[1:]
        t = jc_list[1].threshold[i]
        assert (golden[out_name].iloc[:t] == df[out_name].iloc[:t]).all()
        assert (df[out_name].iloc[t:] == 0).all()
        

    
                

    



"""Tools to work with stock transforms and stock data.
"""

import types
import typing

import pandas

from fast_tx.core import Lag, Accumulate
from fast_tx.utils import date_to_utctimestamp
from fast_tx.stk_tx import (
    TradeFlagOnDiff, StkTradeWithSplitsExecTx, FastTradeCashFlowTx,
    WeightedPositionsTx, FastDivCashFlowTx, FastStkEquityCalc,
    DynamicWeightedPositionsTx)


class NameManager:
    """Class to manage names for stock calculations or other tasks.

    The idea is that you initialize this with a list of string assets
    you are interested in and then you can use it to make names
    as illustrated below:

>>> from fast_tx.stk_tools import NameManager
>>> mngr = NameManager(['SPY', 'GLD'])
>>> mngr.open.n, mngr.close.n
(['SPY_open', 'GLD_open'], ['SPY_close', 'GLD_close'])

    Sub-classes can override the name_for and name_list method if
    desired to control how names are created for assets.
    """

    def __init__(self, asset_list: typing.Sequence[str], specials=(
            ('overall', ('cash', 'cash_deposit', 'div_cash_flow',
                         'equity', 'trade_cash_flow')),)):
        """Initializer.

        :param asset_list:  List of strings reprsenting assets you care about.

        """
        self.asset_list = asset_list
        self.specials = dict(specials)

    @classmethod
    def name_for(cls, element, field, info=None):
        """Utility to generate a name for a field of some element.

        :param element:      Thing we are talking about (e.g., an asset).

        :param field:        Field we want.

        :param info=None:    Extra information

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:  String name.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  The following illustrates example usage:

>>> NameManager.name_for('SPY', 'close')
'SPY_close'
>>> NameManager.name_for('SPY', 'close', 'lagged')
'SPY_close_lagged'

        """
        _ = cls
        result = f'{element}_{field}'
        if info is not None:
            result += '_' + info
        return result

    @classmethod
    def name_list(cls, asset_list, field, info=None):
        "Syntatic sugar to call cls.name_for for everything in asset_list."

        return [cls.name_for(a, field, info) for a in asset_list]

    def __getattr__(self, name):
        result = types.SimpleNamespace()
        if name in self.specials:
            for value in self.specials[name]:
                setattr(result, value, self.name_for(name, value))
        else:
            setattr(result, 'n', self.name_list(self.asset_list, name))
        setattr(self, name, result)  # basically caching this result
        return result


class StkCalcManager:
    """Helper to create transforms to manage stock calculations.
    """

    def __init__(self, names: NameManager):
        """Initializer.

        :param names: NameManager instance to use to generate names.

        """
        self.names = names

    def prep_divs(self, data: pandas.DataFrame, div_pay_dt_offset: int):
        """Prepare canoncial fields required for dividend info.

        :param data:         Dataframe with raw dividend data.

        :param div_pay_dt_offset:  How many days later to pay the dividend.
                                   Setting 0 makes the dividend paid the
                                   same day as the ex-dividend date. This
                                   can be useful for trying to match
                                   adjusted close calculations but is not
                                   realistic so you may want to set longer
                                   offsets.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  Adds div_total and div_pay_dt fields to `data`.

        """
        for i, name in enumerate(self.names.divCash.n):
            divs = data[name]
            div_total = self.names.div_total.n[i]
            div_pay_dt = self.names.div_pay_dt.n[i]
            for ex_date, payment in divs.loc[divs != 0].items():
                iloc = data.index.get_loc(ex_date)
                data.loc[ex_date, div_total] = payment
                data.loc[ex_date, div_pay_dt] = date_to_utctimestamp(
                    data.index[iloc+div_pay_dt_offset].date())

    def make_tx_dict(self, weights, trade_threshold=None, pre_tx_dict=None):
        """Make dictionary of transforms to compute weighted stock positions.

        :param weights:  Either a list of floats for static weights or
                         the string, `'dynamic'` indicating we should look
                         for columns named `self.names.weights.n` for
                         weights.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:  Dictionary of transforms to provide to apply_chain.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  Create a dictionary of transforms to implement the
                  calculations for investing in the given assets based on
                  the given weights applied to the `self.names.overall.equity`
                  field. You will need to prepare a DataFrame with the
                  appropriate data.  See some example usage in
                  the `test_stock_calcs.py` file.

        """
        names = self.names
        if pre_tx_dict:
            tx_list = list(pre_tx_dict.items())
        else:
            tx_list = []
        if trade_threshold:
            tx_list.append(('trade_flag_tx', TradeFlagOnDiff(
                names.position.n, names.close.n, names.post_trade_pos.n,
                names.overall.equity, names.trade_flag.n, trade_threshold)))
        tx_list.extend([
            ('lag_tx', Lag(1, names.next_target.n, names.position.n)),
            ('exec_tx', StkTradeWithSplitsExecTx(
                names.splitFactor.n, names.pre_trade_pos.n, names.position.n,
                names.trade_flag.n, names.post_trade_pos.n)),
            ('trade_cash_flow_tx', FastTradeCashFlowTx(
                names.close.n, names.pre_trade_pos.n, names.post_trade_pos.n,
                names.overall.trade_cash_flow)),
            ('div_cash_flow_tx', FastDivCashFlowTx(
                names.div_total.n, names.div_pay_dt.n, names.pre_trade_pos.n,
                names.overall.div_cash_flow)),
            ('cash_acc_tx', Accumulate([
                names.overall.div_cash_flow, names.overall.trade_cash_flow,
                names.overall.cash_deposit], [names.overall.cash])),
            ('equity_tx', FastStkEquityCalc(
                names.close.n, names.post_trade_pos.n, names.overall.cash,
                names.overall.equity))])
        if weights == 'dynamic':
            tx_list.append(('wt_tx', DynamicWeightedPositionsTx(
                names.weights.n, names.close.n, names.overall.equity,
                names.next_target.n)))
        elif isinstance(weights, list):
            tx_list.append(('wt_tx', WeightedPositionsTx(
                weights, names.close.n, names.overall.equity,
                names.next_position_targets.n)))

        return dict(tx_list)



"""Tools to help manage buffers for fast operations.
"""

import typing

import numba
from numba.typed import List  # pylint: disable=no-member, no-name-in-module
from numba.types import ListType, FunctionType
import numpy
from numpy.typing import NDArray
import pandas



# Define explicit type so we can type hint to numba so it compiles correctly.
CalcRowFuncType = FunctionType(numba.void(
    numba.float64, numba.float64[:], numba.float64[:], numba.float64[:]))
CalcRowFuncListType = ListType(CalcRowFuncType)

# Define explicit type so we can type hint to numba so it compiles correctly.
PostProcRowFuncType = FunctionType(numba.void(
    numba.float64, numba.float64[:], numba.float64[:]))
PostProcRowFuncListType = ListType(PostProcRowFuncType)


@numba.njit(numba.void(numba.float64, numba.float64[:],
                       numba.float64[:]), cache=True)
def noop_post_process_state(index_value: numpy.float64,
                            state: NDArray[numpy.float64],
                            inputs: NDArray[numpy.float64]) -> None:
    """Placeholder for post_process_state method that does nothing.

    Mainly used to be able to compare functions in _apply_inner_loop
    """
    _, _, _ = index_value, state, inputs  # pragma: no cover


def prep_states_array(tx_chain, dataframe):
    """Prepare states array for tx_chain and dataframe.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    :return:  A rectangular numpy array, `states`, where `states[i]`
              is an array for the state for `tx_chain[i]`. Since
              numba requires homogenous input, we must make this rectangular
              even thought `tx_chain[i]` may have a different length state
              than `tx_chain[j]`. We solve this by adding the original size
              of the state array as an extra element to the end so that
              later code can pass the correct sized state to each transform.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    PURPOSE:  Prepare states for tx_chain and dataframe suitable for numba.

    """
    states_list = []
    for t_num, tx in enumerate(tx_chain):
        new_state = tx.startup(dataframe)
        assert new_state is not None, (
            f'startup for {tx} must return (possibly empty) array not None')
        states_list.append(new_state)

    s_len = numpy.asarray([len(s) for s in states_list])
    states = numpy.full((len(tx_chain), max(s_len)), numpy.nan)

    for t_num, tx in enumerate(tx_chain):
        if s_len[t_num]:
            states[t_num, :s_len[t_num]] = states_list[t_num]

    return states, s_len


def prepare_cols(dataframe: pandas.DataFrame,
                 tx_chain: typing.List['FastTx']) -> typing.Tuple[
                     numba.int64[:, :],
                     numba.int64[:, :],
                     typing.Dict[typing.Any, int],
                 ]:
    """Helper for apply_chain to prepare various columns.

    :param dataframe:    DataFrame to work on.

    :param tx_chain:   List of transforms to prepare to run on chain.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    :return:  The tuple (input_cols, output_cols, maincols) such that

         - The array, `input_cols`, is a homogenous numpy array of
           ints with input_cols[t] being the indexes needed by
           tx_chain[t] with input_cols[t][-1] being the number of
           inputs needed for tx_chain[t]. Entries past `input_cols[t][-1]`
           are set to nan. These indexes are relative to the `maincols`
           dict since we will only be working with those columns.
         - The array, `output_cols`, is a homogenous numpy array of
           ints with output_cols[t] being the indexes that tx_chain[t]
           will write to as outputs with output_cols[t][-1] being the number of
           outputs produced by tx_chain[t]. Entries past `output_cols[t][-1]`
           are set to nan. These indexes are relative to the `maincols`
           dict since we will only be working with those columns.
         - The `maincols` dict has keys as column keys (which could be strings
           or more complicated types of column keys) and values are the integer
           index for that column for all the input/output columns needed
           by the transforms in `tx_chain`.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    PURPOSE:   Prepare information for columns as needed by apply_chain.
               See also `_make_buffers` since the columns produced here
               are also needed in there.

    """
    input_cols = []
    output_cols = []
    for tx in tx_chain:
        input_cols.extend(tx.input_col_names())
        output_cols.extend(tx.output_col_names())

    new_cols = set(output_cols) - set(dataframe)
    new_cols = {c: 1 for c in output_cols if c in new_cols}
    dataframe[list(new_cols)] = numpy.nan
    full_cols = list(set(input_cols).union(output_cols))

    maincols = {n: i for i, n in enumerate(full_cols)}

    input_cols = numpy.full((len(tx_chain), 1+max(len(tx.input_col_names())
                                                  for tx in tx_chain)), -1)
    output_cols = numpy.full((len(tx_chain), 1+max(len(tx.output_col_names())
                                                   for tx in tx_chain)), -1)
    for i, tx in enumerate(tx_chain):
        input_cols[i, :len(tx.input_col_names())] = [
            maincols[n] for n in tx.input_col_names()]
        output_cols[i, :len(tx.output_col_names())] = [
            maincols[n] for n in tx.output_col_names()]
        output_cols[i, -1] = -len(tx.output_col_names())

    return input_cols, output_cols, maincols


def prepare_tx_lists(tx_chain, base_func, noop_func):
    """Helper for apply_chain to extract functions to call as lists.

    For numba JIT compilation, we need arrays of functions of the same
    type. This function produces those.
    """
    calc_tx_list = List.empty_list(numba.void(
        numba.float64, numba.float64[:], numba.float64[:], numba.float64[:]
    ).as_type())
    post_process_tx_list = List.empty_list(numba.void(
        numba.float64, numba.float64[:], numba.float64[:]).as_type())

    for tx in tx_chain:
        calc_tx_list.append(tx.calc_row)
        if tx.post_process_state is base_func:
            post_process_tx_list.append(noop_func)
        else:
            post_process_tx_list.append(tx.post_process_state)

    return calc_tx_list, post_process_tx_list


def make_buffers(dataframe, maincols, input_cols, output_cols):
    """Helper for apply_chain to prepare buffers.

    :param dataframe:    Main DataFame to operate on.

    :param maincols:     Iterable of column keys to use for main buffer.

    :param input_cols:   Array of indexes needed for inputs.

    :param output_cols:  Array of indexes needed for outputs.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    :return:  The tuple (index, input_buffer, output_buffer, mainbuffer):

         - The array, `index`, is the dataframe index converted to float
           form if necessary so we can pass to numba jit compliation.
         - The array `input_buffer` is an array of floats big enough to
           to hold the longest input array needed by any transform from
           `input_cols`.
         - The array `output_buffer` is an array of floats big enough to
           to hold the longest output array needed by any transform from
           `output_cols`.
         - The 2-d array, `mainbuffer`, is all the input/output values from
           the main columns of the dataframe that we are going to work on.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    PURPOSE:  We want to prepare `mainbuffer` as just the parts of the
              dataframe we are going to work on. The input_buffer and
              output_buffer are store locations to use for transforms.
              The `index` is a float array suitable for numba. Basically
              we are getting things in a form for fast, efficient, numba
              JIT compiled operation.

    """

    mainbuffer = dataframe[list(maincols)].values

    if isinstance(dataframe.index, pandas.DatetimeIndex):
        index = numpy.array([d.timestamp() for d in dataframe.index])
    else:
        index = dataframe.index.to_numpy()

    input_buffer = numpy.empty(max(len(p) for p in input_cols),
                               dtype=numpy.float64)
    output_buffer = numpy.empty(max(len(p) for p in output_cols),
                                dtype=numpy.float64)

    return index, input_buffer, output_buffer, mainbuffer


@numba.njit(numba.void(numba.float64[:], numba.float64[:, :],
                       numba.float64[:], numba.float64[:],
                       numba.float64[:, :], numba.int64[:],
                       CalcRowFuncListType, PostProcRowFuncListType,
                       numba.int64[:, :], numba.int64[:, :]),
            cache=True)
def apply_inner_loop(
        # pylint: disable=too-many-arguments,too-many-positional-arguments
        index, mainbuffer, input_buffer, output_buffer,
        states, state_len, calc_tx_list, post_process_tx_list,
        input_cols, output_cols):
    """Do inner loop for apply_chain method.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    PURPOSE:  This implements the inner loop for the `apply_chain` function.
              Basically, we take as input the cleaned up and standardized
              versions of everything and then iterate overall all the
              rows of mainbuffer to call `calc_row` for everything and
              then `post_process_state` for everything. This is JIT compiled
              so that looping is efficient but therefore requires all the
              inputs to be relatively standardized types and sizes.

               NOTE: JIT compiling this function is pretty essential.
               NOTE: If you JIT the transforms but not this, its worse
               NOTE: than using no JIT at all.
    """
    for r_num, row in enumerate(mainbuffer):
        for t_num, tx_func in enumerate(calc_tx_list):
            i = 0
            while input_cols[t_num][i] >= 0:
                input_buffer[i] = row[input_cols[t_num, i]]
                i += 1
            tx_func(index[r_num], states[t_num][:state_len[t_num]],
                    input_buffer[:i], output_buffer[:-output_cols[t_num, -1]])
            i = 0
            while output_cols[t_num][i] >= 0:
                row[output_cols[t_num][i]] = output_buffer[i]
                i += 1
        for t_num, tx_func in enumerate(post_process_tx_list):
            if tx_func is noop_post_process_state:
                continue
            i = 0
            while input_cols[t_num][i] >= 0:
                input_buffer[i] = row[input_cols[t_num][i]]
                i += 1
            tx_func(index[r_num], states[t_num][:state_len[t_num]],
                    input_buffer[:i])



"""Test some trivial things.
"""

import numba
import numpy
import pandas
import pytest

from fast_tx.core import (FastTx, Lag, SafeSum, apply_chain)


def test_lookups():

    tx = Lag(1, {'prices': ['a','b','c']}, ['l_a','l_b','l_c'])
    islice = tx.islice('prices')
    assert islice == slice(0, 3)

    # The following verifies code coverage by taking a different
    # path in `islice` via caching.
    assert tx.islice('prices') == slice(0, 3)
    
    with pytest.raises(KeyError) as exc_info:
        tx.islice('unknown')
    assert exc_info.type == KeyError

        



"""This module provides sub-classes of FastTx related to stocks.

These transforms are useful on their own as well as for illustrating
how and why to use the FastTx infrastructure.
"""

import numba
from numba.types import Tuple
import numpy
from numpy.typing import NDArray
import pandas


from fast_tx.core import FastTx


class FastStkEquityCalc(FastTx):
    """Compute equity for stocks (basically value of stocks plus cash).

    The following illustrates example usage:

>>> import pandas
>>> from fast_tx.core import apply_chain
>>> from fast_tx.stk_tx import FastStkEquityCalc
>>> df = pandas.DataFrame({
...    'price_ABC':  [10.00, 10.20, 10.45, 9.80, 11.35],
...    'price_XYZ':  [20.10, 22.22, 21.11, 19.80, 19.35],
...    'shares_ABC': [0, 0, 100, 100, 200],
...    'shares_XYZ': [200, 100, 100, 0, 0]})
>>> df['cash'] = 10000 - (df.shares_ABC.diff() * df.price_ABC) - (
...         df.shares_XYZ.diff() * df.price_XYZ)
>>> df.iloc[0, -1] = 10000
>>> apply_chain(df, [FastStkEquityCalc(
...     ['price_ABC', 'price_XYZ'], ['shares_ABC', 'shares_XYZ'], 'cash',
...     'equity')])
>>> print(df)
   price_ABC  price_XYZ  shares_ABC  shares_XYZ     cash   equity
0      10.00      20.10         0.0       200.0  10000.0  14020.0
1      10.20      22.22         0.0       100.0  12222.0  14444.0
2      10.45      21.11       100.0       100.0   8955.0  12111.0
3       9.80      19.80       100.0         0.0  11980.0  12960.0
4      11.35      19.35       200.0         0.0   8865.0  11135.0

    """

    def __init__(self, close_fields, post_trade_names,
                 cash_name, equity_name):
        """Initializer.

        :param close_fields:   List of names for columns representing closing
                               prices of assets.

        :param post_trade_names:   List of names for columns representing
                                   "post trade" positions. These are basically
                                   the number of shares or units after the
                                   market closes (e.g., after you execute any
                                   trades for the day/bar).

        :param cash_name:          Name of column representing cash owned
                                   after market closes.

        :param equity_name:        Output name representing total equity
                                   after the close. This is basically
                                   the value of cash and stocks and
                                   other assets with equity value (but
                                   not things like futures).

        """
        assert len(close_fields) == len(post_trade_names)

        super().__init__(close_fields + post_trade_names + [cash_name],
                         [equity_name])

    @staticmethod
    @numba.njit(numba.void(numba.float64, numba.float64[:],
                           numba.float64[:], numba.float64[:]), cache=True)
    def calc_row(index_value: numpy.float64, state: NDArray[numpy.float64],
                 inputs: NDArray[numpy.float64],
                 outputs: NDArray[numpy.float64]) -> None:
        """Implement calc_row to compute equity.
        """
        size = len(inputs) >> 1

        start_idx = 0
        closes = inputs[:size]
        start_idx += size
        post_trade_pos = inputs[start_idx:(start_idx+size)]
        equity = inputs[-1]  # inputs[-1] is cash so start with that

        for (close_pos, close_price) in zip(post_trade_pos, closes):
            if not (numpy.isnan(close_pos) or numpy.isnan(close_price)):
                equity += close_pos * close_price

        outputs[0] = equity


class FastTradeCashFlowTx(FastTx):
    """Compute cash flow caused by trading.

>>> import pandas
>>> from fast_tx.core import apply_chain
>>> from fast_tx.stk_tx import FastTradeCashFlowTx
>>> df = pandas.DataFrame({
...     'price_ABC':  [10.0, 10.2, 10.45, 9.8, 11.35],
...     'pre_trade_shares_ABC': [0, 100, 200, 100, 0]})
>>> df['post_trade_shares_ABC'] = df['pre_trade_shares_ABC'].shift(-1)
>>> df.iloc[-1, -1] = 0
>>> apply_chain(df, [FastTradeCashFlowTx(
...     ['price_ABC'], ['pre_trade_shares_ABC'], ['post_trade_shares_ABC'],
...     'cash_flow')])
>>> print(df)  # cash flow is negative when buying and positive when selling
   price_ABC  pre_trade_shares_ABC  post_trade_shares_ABC  cash_flow
0      10.00                   0.0                  100.0    -1000.0
1      10.20                 100.0                  200.0    -1020.0
2      10.45                 200.0                  100.0     1045.0
3       9.80                 100.0                    0.0      980.0
4      11.35                   0.0                    0.0        0.0

    """

    def __init__(self, close_fields, pre_trade_names,
                 post_trade_names, cashflow_name):
        """Initializer.

        :param close_fields:   List of names for columns representing closing
                               prices of assets.

        :param post_trade_names:   List of names for columns representing
                                   "post trade" positions. These are basically
                                   the number of shares or units after the
                                   market closes (e.g., after you execute any
                                   trades for the day/bar).

        :param pre_trade_names:    List of names for columns representing
                                   "pre trade" positions. These are basically
                                   the number of shares or units before the
                                   market opened. The difference in between
                                   values of post_trade_names and
                                   pre_trade_names is interpreted as the
                                   number of units traded. (Splits, stock
                                   dividends or other ways of changing the
                                   share count that do not include trading
                                   should be reflected as a change from
                                   yesterday's post_trade_names to today's
                                   pre_trade_names).

        :param cashflow_name:      Output name representing cash flow
                                   from trading.

        """

        assert len(set(map(len, [
            close_fields, post_trade_names]))) == 1, (
                f'Main input names to {self.__class__} must have same length.')
        self._size = len(close_fields)

        super().__init__(
            close_fields + pre_trade_names + post_trade_names, [cashflow_name])

    def startup(self, dataframe):
        """Slight hack to store number of assets as size of state array.
        """
        return numpy.empty(self._size)

    @staticmethod
    @numba.njit(numba.void(numba.float64, numba.float64[:],
                           numba.float64[:], numba.float64[:]), cache=True)
    def calc_row(index_value: numpy.float64, state: NDArray[numpy.float64],
                 inputs: NDArray[numpy.float64],
                 outputs: NDArray[numpy.float64]) -> None:
        """Implement calc_row to compute cash flow.
        """
        size = len(state)
        start_idx = 0
        closes = inputs[:size]
        start_idx += size
        pre_trade_pos = inputs[start_idx:(start_idx+size)]
        start_idx += size
        post_trade_pos = inputs[start_idx:(start_idx+size)]

        cash_flow = 0
        for i, (close_pos, close_price) in enumerate(zip(
                post_trade_pos, closes)):
            if not (numpy.isnan(close_pos) or numpy.isnan(close_price)):
                pre_pos = 0 if numpy.isnan(pre_trade_pos[i]) else (
                    pre_trade_pos[i])
                bought = close_pos - pre_pos
                cash_flow -= close_price * bought

        assert not numpy.isnan(cash_flow)

        outputs[0] = cash_flow


class FastDivCashFlowTx(FastTx):
    """Compute cash flow from dividends.

    See also the docs for the calc_row and post_process_state methods
    since there are some subtelties to this calculation.

The following illustrates example usage.

>>> import pandas, numpy
>>> from fast_tx.core import apply_chain
>>> from fast_tx.utils import date_to_utctimestamp
>>> from fast_tx.stk_tx import FastDivCashFlowTx
>>> df = pandas.DataFrame({
...     'shares_ABC': [100, 200, 100, 0, 0],
...     'div_ABC': [0, 0, .2, 0, 0],
...     'pay_div_date_ABC': [numpy.nan, numpy.nan,
...             pandas.Timestamp('2025-05-23'), numpy.nan, numpy.nan]},
...     index=[pandas.Timestamp(p)
...     for p in [f'2025-05-{i}' for i in range(19, 24)]])
>>> print(df)
            shares_ABC  div_ABC pay_div_date_ABC
2025-05-19         100      0.0              NaT
2025-05-20         200      0.0              NaT
2025-05-21         100      0.2       2025-05-23
2025-05-22           0      0.0              NaT
2025-05-23           0      0.0              NaT
>>> # Above represents case where ex-dividend date is 2025-05-21,
>>> # and dividend amount is 20 cents per share to be paid on 2025-05-23.
>>> df['pay_div_date_ABC'] = [date_to_utctimestamp(d)  # convert date column
...     for d in df['pay_div_date_ABC']]               # to float for numba
>>> apply_chain(df, [FastDivCashFlowTx(
...     ['div_ABC'], ['pay_div_date_ABC'], ['shares_ABC'],
...     'div_cash_flow_ABC')])
>>> print(df)
            shares_ABC  div_ABC  pay_div_date_ABC  div_cash_flow_ABC
2025-05-19       100.0      0.0               NaN                0.0
2025-05-20       200.0      0.0               NaN                0.0
2025-05-21       100.0      0.2      1.747958e+09                0.0
2025-05-22         0.0      0.0               NaN                0.0
2025-05-23         0.0      0.0               NaN               40.0

Note that the dividend of 20 cents per share which is ex-dividend on
2025-05-21 is applied to the 200 shares owned immediately **BEFORE**
the ex-dividend date for a total payment of $40 received on the payment
date of 2025-05-23.

    """

    def __init__(self, div_totals, div_pay_dts,
                 post_trade_names, cashflow_name):
        """Initializer.

        :param div_totals:   List of columns indicating the total dividends
                             on a given record date.

        :param div_pay_dts:  List of columns indicating the payment date
                             for a dividend on a given ex-dividend date.

        :param post_trade_names:  List of columns showing how much stock
                                  is owned at the close. We will use this
                                  value on the day **BEFORE** the ex-dividend
                                  date to determine how many shares to use
                                  in computing the dividend.

        :param cashflow_name:     Output name for the total dividend cash flow.

        """
        assert len(set(map(len, [
            div_totals, div_pay_dts, post_trade_names]))) == 1, (
                f'Main input names to {self.__class__} must have same length.')
        self._size = len(div_totals)

        super().__init__(
            div_totals + div_pay_dts + post_trade_names, [cashflow_name])

    def startup(self, dataframe):
        result = numpy.full(3*self._size + 1, numpy.nan)
        result[-1] = self._size
        return result

    @staticmethod
    @numba.njit(numba.void(numba.float64, numba.float64[:],
                           numba.float64[:], numba.float64[:]), cache=True)
    def calc_row(index_value: numpy.float64, state: NDArray[numpy.float64],
                 inputs: NDArray[numpy.float64],
                 outputs: NDArray[numpy.float64]) -> None:
        """Implement calculate row.

        If there is a dividend, we will want to know the payment date
        and save that payment date in the state so we can pay out the
        dividend on that date. We also need to know how many shares
        to use to compute the total dividend (see below).

        The post_process_state method will look at the post_trade_pos
        after all information for a row has been processed and put that into
        the state so this method can look at the post_trade_pos from the
        previous row. That is the amount of shares we owned **BEFORE** the
        ex-dividend date and can be used to determine the dividend amount.

        So this function first checks if there is a dividend to remember
        and then checks if there are dividends remembered to pay out.
        """
        size = numpy.int64(state[-1])
        start_idx = 0
        divs = inputs[start_idx:(size+start_idx)]
        pending_div_vals = state[start_idx:(size+start_idx)]
        start_idx += size
        div_dates = inputs[start_idx:(size+start_idx)]
        pending_div_dates = state[start_idx:(size+start_idx)]
        start_idx += size
        prev_pos = state[start_idx:(size+start_idx)]

        cash_flow = 0
        for i in range(size):  # Check if we have non-zero div + previous pos
            if not (numpy.isnan(divs[i]) or divs[i] == 0
                    or numpy.isnan(prev_pos[i])):
                assert numpy.isnan(pending_div_dates[i]), (
                    'Cannot have more than 1 pending dividend')
                pending_div_dates[i] = div_dates[i]    # Save div in state
                pending_div_vals[i] = numpy.floor(     # to check in future.
                    100*divs[i]*prev_pos[i])/100.0     # Drop fractional cents.

            if index_value >= pending_div_dates[i]:  # div received now
                cash_flow += pending_div_vals[i]
                pending_div_dates[i] = numpy.nan  # this will clear state since
                pending_div_vals[i] = numpy.nan  # pending div is processed

        assert not numpy.isnan(cash_flow)
        outputs[0] = cash_flow

    @staticmethod
    @numba.njit(numba.void(numba.float64, numba.float64[:],
                           numba.float64[:]), cache=True)
    def post_process_state(index_value, state, inputs):
        """Post process the row to determine the state.

        We have to update the state **AFTER** the row has been fully
        processed so we can see the post trade position. We will then
        save the post trade position for this row into the state so that
        when `calc_row` is executed on the next row it can look at the
        state to get the previous row's post trade position.
        """
        size = numpy.int64(state[-1])
        sizex2 = size << 1
        sizex3 = sizex2 + size
        post_trade_pos = inputs[sizex2:sizex3]
        prev_pos = state[sizex2:sizex3]

        for i in range(size):
            prev_pos[i] = post_trade_pos[i]  # save post trade pos in state


class StkTradeExecTx(FastTx):
    """Compute calculating stock trade executions.

    See StkTradeWithSplitsExecTx if you need to handle stock splits.

    Basically this takes in some target positions and trade flags and
    produces the pre and post trade positions. This is useful for when
    you have a quantitative strategy computing target positions all the
    time but you only want to actually trade occaionsally (e.g., every
    Tuesday or every month end or when target positions are far enough
    from current positions).

    The following illustrates example usage:

>>> import pandas
>>> from fast_tx.core import apply_chain
>>> from fast_tx.stk_tx import StkTradeExecTx
>>> df = pandas.DataFrame({
...     'target_shares':  [100*i for i in range(6)],
...     'trade_flag': [0, 1, 0, 0, 1, 0]})
>>> apply_chain(df, [StkTradeExecTx(
...     ['pre_trade_shares'], ['target_shares'], ['trade_flag'],
...     ['post_trade_shares'])])
>>> print(df)  # cash flow is negative when buying and positive when selling
   target_shares  trade_flag  pre_trade_shares  post_trade_shares
0            0.0         0.0               0.0                0.0
1          100.0         1.0               0.0              100.0
2          200.0         0.0             100.0              100.0
3          300.0         0.0             100.0              100.0
4          400.0         1.0             100.0              400.0
5          500.0         0.0             400.0              400.0

Notice that while the target shares is changing each day, we only
actually trade on days indicated by `trade_flag`.

    """
    def __init__(self, pre_trade_names, target_names, trade_flag_names,
                 post_trade_names):
        """Initializer.

        :param pre_trade_names:    List of names for columns representing
                                   "pre trade" positions. These are basically
                                   the number of shares or units before the
                                   market opened. The difference in between
                                   values of post_trade_names and
                                   pre_trade_names is interpreted as the
                                   number of units traded. (Splits, stock
                                   dividends or other ways of changing the
                                   share count that do not include trading
                                   should be reflected as a change from
                                   yesterday's post_trade_names to today's
                                   pre_trade_names).


        :param target_names:       List of names indicating the target
                                   position to adopt when indicated by
                                   trade flag.

        :param trade_flag_names:   List of names indicating when to trade.

        :param post_trade_names:   List of names for columns representing
                                   "post trade" positions. These are basically
                                   the number of shares or units after the
                                   market closes (e.g., after you execute any
                                   trades for the day/bar).

        """
        assert len(pre_trade_names) == len(post_trade_names)
        assert len(pre_trade_names) == len(trade_flag_names)
        assert len(target_names) == len(pre_trade_names)
        super().__init__(target_names + trade_flag_names,
                         pre_trade_names + post_trade_names)

    def startup(self, dataframe: pandas.DataFrame) -> NDArray[numpy.float64]:
        out_size = len(self.output_col_names())
        result = numpy.zeros(1 + out_size//2)
        result[0] = out_size//2
        return result

    @staticmethod
    @numba.njit(numba.void(numba.float64, numba.float64[:],
                           numba.float64[:], numba.float64[:]), cache=True)
    def calc_row(index_value: numpy.float64, state: NDArray[numpy.float64],
                 inputs: NDArray[numpy.float64],
                 outputs: NDArray[numpy.float64]) -> None:
        size = numpy.int32(state[0])
        targets = inputs[:size]
        flags = inputs[size:]
        for i, f in enumerate(flags):
            outputs[i] = state[i+1]  # set pre trade pos based on previous day
            if f:  # trade flag says to trade
                outputs[i+size] = targets[i]  # set post trade position to tgt
            else:
                outputs[i+size] = outputs[i]  # set post trade position to pre
            state[i+1] = outputs[i+size]  # save post trade pos in state


class WeightedPositionsTx(FastTx):
    """Take weighted positions in assets based on current equity.

    The following illustrates example usage:

>>> import pandas, numpy
>>> from fast_tx.core import apply_chain
>>> from fast_tx.stk_tx import (
...     WeightedPositionsTx, FastStkEquityCalc, StkTradeExecTx)
>>> df = pandas.DataFrame({
...    'price_ABC':  [10, 10, 10, 9, 11],
...    'price_XYZ':  [20, 22, 21, 19, 19],
...    'trade_flag_ABC': [0, 1, 0, 1, 0],
...    'trade_flag_XYZ': [0, 1, 0, 1, 0],
...    'shares_ABC': [0, 0, 100, 100, 200],
...    'shares_XYZ': [200, 100, 100, 0, 0]})
>>> df['cash'] = 10000
>>> df.iloc[1:, -1] = 10000 - ((df.shares_ABC.diff() * df.price_ABC) - (
...         df.shares_XYZ.diff() * df.price_XYZ)).iloc[1:]
>>> apply_chain(df, [FastStkEquityCalc(
...     ['price_ABC', 'price_XYZ'], ['shares_ABC', 'shares_XYZ'], 'cash',
...     'equity'), WeightedPositionsTx([.25, .75],
...     ['price_ABC', 'price_XYZ'], 'equity', ['target_ABC', 'target_XYZ']),
...     StkTradeExecTx(['pre_trd_ABC','pre_trd_XYZ'],
...         ['target_ABC', 'target_XYZ'], ['trade_flag_ABC', 'trade_flag_XYZ'],
...         ['post_trd_ABC','post_trd_XYZ'])])
>>> print(df.iloc[:,10:])
   pre_trd_ABC  pre_trd_XYZ  post_trd_ABC  post_trd_XYZ
0          0.0          0.0           0.0           0.0
1          0.0          0.0         250.0         340.0
2        250.0        340.0         250.0         340.0
3        250.0        340.0         250.0         355.0
4        250.0        355.0         250.0         355.0

    """

    def __init__(self, weights, price_names, equity_name, position_names):
        """Initializer.

        :param weights:      Names indicating weights for each asset.

        :param price_names:  Names indicating prices for each asset.

        :param equity_name:  Name indicating total equity.

        :param position_names:  Target positions to compute.

        """
        assert len(weights) == len(position_names)
        assert len(weights) == len(price_names)
        self._weights = weights
        super().__init__(price_names + [equity_name], position_names)

    def startup(self, dataframe: pandas.DataFrame) -> NDArray[numpy.float64]:
        return numpy.array(self._weights, dtype=numpy.float64)

    @staticmethod
    @numba.njit(numba.void(numba.float64, numba.float64[:],
                           numba.float64[:], numba.float64[:]), cache=True)
    def calc_row(index_value: numpy.float64, state: NDArray[numpy.float64],
                 inputs: NDArray[numpy.float64],
                 outputs: NDArray[numpy.float64]) -> None:
        equity = inputs[-1]
        for i, weight in enumerate(state):
            outputs[i] = numpy.floor(weight * equity / inputs[i])


class DynamicWeightedPositionsTx(FastTx):
    """Take weighted positions in assets based on current equity.

    The following illustrates example usage:

>>> import pandas, numpy
>>> from fast_tx.core import apply_chain
>>> from fast_tx.stk_tx import (
...     DynamicWeightedPositionsTx, FastStkEquityCalc, StkTradeExecTx)
>>> df = pandas.DataFrame({
...    'price_ABC':  [10, 10, 10, 9, 11],
...    'price_XYZ':  [20, 22, 21, 19, 19],
...    'weight_ABC':  [.1, .1, .5, .9, .9],
...    'weight_XYZ':  [.9, .9, .5, .1, .1],
...    'trade_flag_ABC': [0, 1, 0, 1, 0],
...    'trade_flag_XYZ': [0, 1, 0, 1, 0],
...    'shares_ABC': [0, 0, 100, 100, 200],
...    'shares_XYZ': [200, 100, 100, 0, 0]})
>>> df['cash'] = 10000
>>> df.iloc[1:, -1] = 10000 - ((df.shares_ABC.diff() * df.price_ABC) - (
...         df.shares_XYZ.diff() * df.price_XYZ)).iloc[1:]
>>> apply_chain(df, [FastStkEquityCalc(
...     ['price_ABC', 'price_XYZ'], ['shares_ABC', 'shares_XYZ'], 'cash',
...     'equity'), DynamicWeightedPositionsTx(['weight_ABC', 'weight_XYZ'],
...     ['price_ABC', 'price_XYZ'], 'equity', ['target_ABC', 'target_XYZ']),
...     StkTradeExecTx(['pre_trd_ABC','pre_trd_XYZ'],
...         ['target_ABC', 'target_XYZ'], ['trade_flag_ABC', 'trade_flag_XYZ'],
...         ['post_trd_ABC','post_trd_XYZ'])])
>>> df['ratio'] = df.price_ABC*df.target_ABC/df.price_XYZ/df.target_XYZ
>>> print(df.iloc[:,12:])
   pre_trd_ABC  pre_trd_XYZ  post_trd_ABC  post_trd_XYZ     ratio
0          0.0          0.0           0.0           0.0  0.111111
1          0.0          0.0         100.0         409.0  0.111136
2        100.0        409.0         100.0         409.0  1.000331
3        100.0        409.0         900.0          47.0  9.070549
4        900.0         47.0         900.0          47.0  9.063521

    """

    def __init__(self, weight_names, price_names, equity_name, position_names):
        """Initializer.

        :param weight_names:  Names indicating weights for each asset.

        :param price_names:   Names indicating prices for each asset.

        :param equity_name:   Name indicating total equity.

        :param position_names:  Target positions to compute.

        """
        assert len(weight_names) == len(position_names)
        assert len(weight_names) == len(price_names)
        super().__init__(price_names + weight_names + [equity_name],
                         position_names)

    @staticmethod
    @numba.njit(numba.void(numba.float64, numba.float64[:],
                           numba.float64[:], numba.float64[:]), cache=True)
    def calc_row(index_value: numpy.float64, state: NDArray[numpy.float64],
                 inputs: NDArray[numpy.float64],
                 outputs: NDArray[numpy.float64]) -> None:
        size = len(outputs)
        double_size = size << 1
        weights = inputs[size:double_size]
        equity = inputs[-1]
        for i, price in enumerate(inputs[:size]):
            outputs[i] = numpy.floor(weights[i] * equity / price)


class CashToFracSharesTx(FastTx):
    """Use cash to buy shares and include fractional shares.
    """

    def __init__(self, price_names, cash_names, position_names):
        assert len(price_names) == len(position_names)
        super().__init__(price_names + cash_names, position_names)

    def startup(self, dataframe):
        state = numpy.zeros(len(self.output_col_names()))
        return state

    @staticmethod
    @numba.njit(numba.void(numba.float64, numba.float64[:],
                           numba.float64[:], numba.float64[:]), cache=True)
    def calc_row(index_value: numpy.float64, state: NDArray[numpy.float64],
                 inputs: NDArray[numpy.float64],
                 outputs: NDArray[numpy.float64]) -> None:
        size = len(outputs)
        cash = sum(inputs[size:])
        for i, price in enumerate(inputs[:size]):
            outputs[i] = state[i] + cash / price / size
            state[i] = outputs[i]


class StkTradeWithSplitsExecTx(FastTx):
    """Compute calculating stock trade executions including stock splits.

    Similar to StkTradeExecTx but handles splits.

    Basically this takes in some target positions and trade flags and
    produces the pre and post trade positions while taking account
    possible stock splits (and reverse splits). This is useful for when
    you have a quantitative strategy computing target positions all the
    time but you only want to actually trade occaionsally (e.g., every
    Tuesday or every month end or when target positions are far enough
    from current positions).

    The following illustrates example usage:

>>> import pandas
>>> from fast_tx.core import apply_chain
>>> from fast_tx.stk_tx import StkTradeWithSplitsExecTx
>>> df = pandas.DataFrame({
...     'target_shares':  [100*i for i in range(6)],
...     'split_factor': [1, 1, .5, 1, 2, 1],
...     'trade_flag': [0, 1, 0, 0, 1, 0]})
>>> apply_chain(df, [StkTradeWithSplitsExecTx(
...     ['split_factor'], ['pre_trade_shares'], ['target_shares'],
...     ['trade_flag'], ['post_trade_shares'])])
>>> print(df)  # cash flow is negative when buying and positive when selling
   target_shares  split_factor  trade_flag  pre_trade_shares  post_trade_shares
0            0.0           1.0         0.0               0.0                0.0
1          100.0           1.0         1.0               0.0              100.0
2          200.0           0.5         0.0              50.0               50.0
3          300.0           1.0         0.0              50.0               50.0
4          400.0           2.0         1.0             100.0              400.0
5          500.0           1.0         0.0             400.0              400.0

Notice that while the target shares is changing each day, we only
actually trade on days indicated by `trade_flag`.

    """
    def __init__(self, split_factor_names, pre_trade_names, target_names,
                 trade_flag_names, post_trade_names):
        """Initializer.

        :param split_factor_names:  List of names for columns representing
                                    "split factor". If this is less than
                                    1 it means a reverse split and if it
                                    is greater than 1 it means a split. We
                                    set the pre trade position to the
                                    split factor times the previous post
                                    trade position.

        :param pre_trade_names:    List of names for columns representing
                                   "pre trade" positions. These are basically
                                   the number of shares or units before the
                                   market opened. The difference in between
                                   values of post_trade_names and
                                   pre_trade_names is interpreted as the
                                   number of units traded. (Splits, stock
                                   dividends or other ways of changing the
                                   share count that do not include trading
                                   should be reflected as a change from
                                   yesterday's post_trade_names to today's
                                   pre_trade_names).


        :param target_names:       List of names indicating the target
                                   position to adopt when indicated by
                                   trade flag.

        :param trade_flag_names:   List of names indicating when to trade.

        :param post_trade_names:   List of names for columns representing
                                   "post trade" positions. These are basically
                                   the number of shares or units after the
                                   market closes (e.g., after you execute any
                                   trades for the day/bar).

        """
        assert len(split_factor_names) == len(pre_trade_names)
        assert len(pre_trade_names) == len(post_trade_names)
        assert len(pre_trade_names) == len(trade_flag_names)
        assert len(target_names) == len(pre_trade_names)
        super().__init__(target_names + trade_flag_names + split_factor_names,
                         pre_trade_names + post_trade_names)

    def startup(self, dataframe: pandas.DataFrame) -> NDArray[numpy.float64]:
        out_size = len(self.output_col_names())
        result = numpy.zeros(1 + out_size//2)
        result[0] = out_size//2
        return result

    @staticmethod
    @numba.njit(numba.void(numba.float64, numba.float64[:],
                           numba.float64[:], numba.float64[:]), cache=True)
    def calc_row(index_value: numpy.float64, state: NDArray[numpy.float64],
                 inputs: NDArray[numpy.float64],
                 outputs: NDArray[numpy.float64]) -> None:
        size = numpy.int32(state[0])
        double_size = size << 1
        targets = inputs[:size]
        flags = inputs[size:double_size]
        split_factors = inputs[double_size:]
        for i, f in enumerate(flags):
            # set pre trade pos based on previous day
            outputs[i] = state[i+1] * split_factors[i]
            if f:  # trade flag says to trade
                outputs[i+size] = targets[i]  # set post trade position to tgt
            else:
                outputs[i+size] = outputs[i]  # set post trade position to pre
            state[i+1] = outputs[i+size]  # save post trade pos in state


class TradeFlagOnDiff(FastTx):
    """Generate trade flags based on difference in positions.

This transform is designed to generate a trade flag (value = 1) if
the sum of absolute differences between the target and post_trade
positions (weighted by price) exceeds some threshold of equity.

The following illustrates example usage:

>>> import pandas
>>> from fast_tx.core import apply_chain
>>> from fast_tx.stk_tx import TradeFlagOnDiff, FastStkEquityCalc
>>> df = pandas.DataFrame({
...     't_ABC':  [100*i + 1 for i in range(6)],
...     't_XYZ':  [200*i + 1 for i in range(6)],
...     'pos_ABC':  [100*i for i in range(6)],
...     'pos_XYZ':  [200*i for i in range(6)],
...     'price_ABC':  [(10 + i*.1) for i in range(6)],
...     'price_XYZ':  [(20 - i*.15) for i in range(6)],
...     'cash': [100]*6})
>>> df.loc[4,'price_XYZ'] = 2
>>> df.loc[4,'pos_XYZ'] = 700
>>> apply_chain(df, [FastStkEquityCalc(['price_ABC', 'price_XYZ'],
...     ['pos_ABC', 'pos_XYZ'], 'cash', 'equity'),
... TradeFlagOnDiff(
...     ['t_ABC', 't_XYZ'], ['price_ABC', 'price_XYZ'],
...     ['pos_ABC', 'pos_XYZ'], 'equity',
...     ['flag_ABC', 'flag_XYZ'])])
>>> print(df[[n for n in df if n not in ('cash','equity')]])
   t_ABC   t_XYZ  pos_ABC  pos_XYZ  price_ABC  price_XYZ  flag_ABC  flag_XYZ
0    1.0     1.0      0.0      0.0       10.0      20.00       0.0       0.0
1  101.0   201.0    100.0    200.0       10.1      19.85       1.0       1.0
2  201.0   401.0    200.0    400.0       10.2      19.70       0.0       0.0
3  301.0   601.0    300.0    600.0       10.3      19.55       0.0       0.0
4  401.0   801.0    400.0    700.0       10.4       2.00       0.0       0.0
5  501.0  1001.0    500.0   1000.0       10.5      19.25       1.0       1.0

Notice that we have a trade triggered at index 1 because of the big difference
between position and target on index 0 and again on index 5 because of the big
difference between position and target on index 5.

    """
    def __init__(self, target_names, price_names, post_trade_names,
                 equity_name, trade_flag_names, threshold=0.01):
        """Initializer.

        :param target_names:       List of names indicating the target
                                   position to adopt if we trade.

        :param price_names:        Prices corresponding to target_names

        :param post_trade_names:   List of names indicating post trade
                                   positions.

        :param equity:             String name indicating equity column.

        :param trade_flag_names:   List of names indicating when to trade.
                                   If these exist prior to applying this
                                   transform, we "OR" the operation of this
                                   transform with existing flags. This lets
                                   do something like let you set trade flags
                                   for the start of the month yourself and
                                   then apply this class to get threshold
                                   based trades in addition.

        :param threshold:          Minimum threshold above which trade flags
                                   are turned on.
        """
        assert len(price_names) == len(target_names)
        assert len(price_names) == len(post_trade_names)
        assert len(price_names) == len(trade_flag_names)
        assert threshold > 0
        self._threshold = threshold
        super().__init__({'target': target_names, 'price': price_names,
                          'post_trade': post_trade_names,
                          'trade_flags': trade_flag_names,
                          'equity': [equity_name]}, trade_flag_names)
        self.bind_unpack()
        self.bind_post_process_state()
        self.bind_calc_row()

    def startup(self, dataframe: pandas.DataFrame) -> NDArray[numpy.float64]:
        """Implement as required by parent.

We return a buffer with the same size as the output trade flags.  It
will be used by post_process_state to capture the intended value of
the output trade flags. See bind_post_process_state and
post_process_state for details of how the state works.
        """
        out_size = len(self.output_col_names())
        result = numpy.zeros(out_size)
        return result

    def bind_calc_row(self):
        """Bind a function for the calc_row method.

We dynamically bind the calc_row method as a closure so the numba
compiled version can easily see the size and sizex4 variables.
        """
        size = len(self.output_col_names())
        flag_slice = self.islice('trade_flags')
        @numba.njit(numba.void(numba.float64, numba.float64[:],
                               numba.float64[:], numba.float64[:]),
                    cache=True)
        def calc_row(_index_value: numpy.float64,
                     state: NDArray[numpy.float64],
                     inputs: NDArray[numpy.float64],
                     outputs: NDArray[numpy.float64]) -> None:
            """Calculte the trade flags for the row.

We set the output trade flags based on the flag values saved to the state
by post_process_state combined with the logical AND of the trade flag
values seen as input. See docs of post_process_state for more details.
            """
            flags = inputs[flag_slice]
            for i in range(size):
                outputs[i] = state[i] or (  # also check if flag is
                    (not numpy.isnan(flags[i])) and flags[i])  # already on
        self.calc_row = calc_row

    def bind_unpack(self):
        """Dynamacially bind the unpack method.
        """
        threshold = self._threshold
        size = len(self.output_col_names())  # used in closure below
        target_slice = self.islice('target')
        price_slice = self.islice('price')
        post_trade_slice = self.islice('post_trade')
        equity_idx = self.index('equity')

        @numba.njit(Tuple((
            numba.float64, numba.float64, numba.float64,
            numba.float64[:], numba.float64[:], numba.float64[:]))(
                numba.float64[:], numba.float64[:]), cache=True)
        def unpack(_state, inputs):
            """Helper to unpack state and inputs into needed variables.
            """
            equity = inputs[equity_idx]
            targets = inputs[target_slice]
            prices = inputs[price_slice]
            post_trade = inputs[post_trade_slice]
            return (threshold, size, equity, targets, prices, post_trade)
        self.unpack = unpack

    def bind_post_process_state(self):
        """Dynamically bind the post_process_state method.
        """
        unpack = self.unpack

        @numba.njit(numba.void(numba.float64, numba.float64[:],
                               numba.float64[:]), cache=True)
        def post_process_state(_index_value, state, inputs):
            """Post process row to get post trade positions, equity, etc..

We update state *AFTER* row has been fully processed so we can see the
post trade position, equity, and so on. We then save post trade
position for row into state so when `calc_row` is executed on next
row, it can look at the state to get the previous row's post trade
position.
            """
            (threshold, size, equity, targets, prices, post_trade
             ) = unpack(state, inputs)
            abs_diff = 0
            for i in range(size):
                abs_diff += abs(
                    (0 if numpy.isnan(targets[i]) else targets[i]
                     ) - (0 if numpy.isnan(post_trade[i]) else post_trade[i])
                ) * (0 if numpy.isnan(prices[i]) else prices[i])
            if equity != 0 and abs_diff/equity > threshold:
                for i in range(size):
                    state[i] = 1
            else:
                for i in range(size):
                    state[i] = 0
        self.post_process_state = post_process_state



"""Tests for stock related calculations.
"""

import cProfile
import io
import pstats
from pstats import SortKey

from pathlib import Path

import numpy
import pandas

import fast_tx_tests
from fast_tx.core import apply_chain
from fast_tx.utils import date_to_utctimestamp
from fast_tx import stk_tools
from fast_tx.stk_tx import CashToFracSharesTx
from fast_tx.stk_tools import NameManager

name_for = NameManager.name_for
name_list = NameManager.name_list


def tpath() -> str:
    "Get test path."
    return fast_tx_tests.__path__[0]  # type: ignore


def load_stock_file(asset):
    """Load stock data for a specified asset from a compressed CSV file.

    :param asset: The name or identifier of the asset for which to load data.

    :return: A pandas DataFrame containing the loaded stock data.

    PURPOSE: Load and label stock data from a compressed CSV file,
             parsing the index as dates and renaming columns using a
             predefined naming convention.
    """
    fname = Path(tpath()) / 'data' / f'test_data_{asset}.csv'
    data = pandas.read_csv(fname, index_col=0, parse_dates=True)
    data.columns = [name_for(asset, n) for n in data.columns]
    return data


def prep_position(data, asset, pos_locs=((.2, .4, 1), (.6, .8, 1))):
    """Prepare position data for specified asset.

    :param data:      DataFrame to modify.
    :param asset:     Asset name for labeling.
    :param pos_locs:  Tuple of (start, end, val) locations for positions.
                      Each window is set to have position be `val` between
                      the start and end points (interpreted as floating
                      point fractions of the date index).

    PURPOSE: Set positions in data for given asset based on pos_locs.
    """
    data[name_for(element=asset, field='position')] = 0
    for start_pos_loc, end_pos_loc, val in pos_locs:
        if isinstance(start_pos_loc, float):
            start_pos_loc = int(round(len(data.index) * start_pos_loc))
        if isinstance(end_pos_loc, float):
            end_pos_loc = int(round(len(data.index) * end_pos_loc))
        data.loc[data.index[start_pos_loc]:data.index[end_pos_loc],
                 name_for(element=asset, field='position')] = val

    # Fast calculations will generally want float arrays so we
    # convert bool to float.
    data[name_for(element=asset, field='trade_flag')] = [
        float(d.weekday() == 2) for d in data.index]


def prep_full_stock_pnl(asset_list, div_pay_dt_offset=1):
    """Prepare and process stock data for profit and loss calculations.

    :param asset_list: List of asset names or identifiers to process.

    :return: DataFrame with processed stock data.

    PURPOSE: Load, label, and prepare stock data for PnL calculations.
    """
    data = pandas.concat([load_stock_file(asset) for asset in asset_list],
                         axis=1)
    stk_tools.StkCalcManager(stk_tools.NameManager(asset_list)).prep_divs(
        data, div_pay_dt_offset=div_pay_dt_offset)
    for asset in asset_list:
        prep_position(data, asset)
    return data


def check_against_golden_file(data, golden_file, check_fields, tol=1e-6):
    """Check that data matches golden_file

    :param data:   DataFrame to check.

    :param golden_file:   Optional path to golden file.

    :param check_fields:   List of columns to check.

    :param tol:  Tolerance for differences.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    :return:  DataFrame showing locations where differences exceed tol.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    PURPOSE:  If `golden_file` is provided, check that it matches `data`.

    """
    diff = None
    if golden_file:
        golden = pandas.read_csv(golden_file, index_col=0, parse_dates=True)
        diff = data[check_fields].loc[data[check_fields].sub(
            golden[check_fields]).abs().sum(axis=1) > tol]
        assert diff.empty, f'Too many diffs vs golden:\n{diff}'

    return diff


def make_stk_prof_tx_dict(asset_list, weights, **kwargs):
    """Make dictionary of transforms for stock calculations.

    :param asset_list:    List of assets to work on.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    :return:  Dictionary of transforms for stock calculations.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    PURPOSE:  Create a dictionary of transforms to do stock calculations
              so we can use these in testing.

    NOTE:  If you really wanted to handle dividends correctly, you
           would need to deal with a number of additional
           complications. One issues is that when stock settlement
           went from T+2 to T+1 the ex-date became the same as the
           record-date.  For example, see the URL below (it is just
           one URL split across 3 lines; just combine all the lines:

                 https://web.archive.org/web/20250611200750/\
                     https://www.sifma.org/wp-content/uploads/2024/05/\
                        SIFMA-T1-Dividend-Processing-FAQ.pdf

    """
    names = NameManager(asset_list)
    stk_mngr = stk_tools.StkCalcManager(names)
    tx_dict = stk_mngr.make_tx_dict(weights=weights, **kwargs)
    return tx_dict


def make_full_stock_pnl_info(asset_list, div_pay_dt_offset=1,
                             start_cash=10000):
    """Prepare DataFrame for stock profit and loss calculations.

    :param asset_list:  List of assets to include.

    :param div_pay_dt_offset=1:  How many days after the dividend record date
                                 the dividend will be received.

    :param start_cash=10000:     How much cash to start with.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    :return:  DataFrame with test data prepared.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    PURPOSE:  Prepare data for testing.

    """
    data = prep_full_stock_pnl(asset_list, div_pay_dt_offset=div_pay_dt_offset)
    data.drop(columns=name_list(asset_list, 'position'), inplace=True)

    cash_dep_name = name_for(field='cash_deposit', element='overall')
    data[cash_dep_name] = 0
    data.loc[data.index[int(len(data.index)*.05)], cash_dep_name] = start_cash
    data[name_list(asset_list, 'trade_flag')] = numpy.float64(
        data[name_list(asset_list, 'trade_flag')])
    for name in name_list(asset_list, 'div_pay_dt'):
        data[name] = [d if isinstance(d, float) else date_to_utctimestamp(d)
                      for d in data[name]]

    return data


def check_fixme(asset_list, golden_file, start_cash=100000000, **kwargs):
    """Verify equity calculations match adjusted close.

    :param asset:    Single asset to check (e.g., `'SPY'`).

    :param golden_file:   Optional path to golden file to check.

    :param start_cash=100000000:   Amount of starting cash.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    :return:  The tuple (data, diff).

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    PURPOSE:  This function puts together a DataFrame for the given `asset`
              where we immediately invest any cash deposit or dividend back
              into the asset. This is kind of cheating because:

                 1. In reality, you would not immediately receive until later.
                 2. You may not actually be able to invest all your cash
                    exactly at the current close price because you would not
                    know the close price beforehand and so would not know how
                    to setup the shares/price of a market/limit order.

              Still cheating about the above is a way to verify our
              calculations against a vendor provided adjusted close.
    """
    data = make_full_stock_pnl_info(
        asset_list, div_pay_dt_offset=1, start_cash=start_cash)

    if kwargs.get('trade_threshold', None) or kwargs.get(
            'pre_tx_dict', {}).get('trade_flag_tx', None):  # type: ignore
        # trading will be controled by threshold or transform
        data[name_list(asset_list, 'trade_flag')] = numpy.nan
    else:  # no threshold so turn on trade flag
        data[name_list(asset_list, 'trade_flag')] = 1.0
    size = float(len(asset_list))/2.0 * (len(asset_list) + 1)
    for i, name in enumerate(name_list(asset_list, 'weights')):
        data[name] = .99*(i+1.0)/size
        data.loc[data.index[-100]:, name] = .99*(len(asset_list)-i)/size

    orig_tx_dict = make_stk_prof_tx_dict(asset_list, weights='dynamic',
                                         **kwargs)
    apply_chain(data, list(orig_tx_dict.values()))

    diff = check_against_golden_file(data, golden_file, (
        list(set(orig_tx_dict['exec_tx'].input_col_names() +
                 orig_tx_dict['exec_tx'].output_col_names())
             ) + [name_for('overall', 'equity')]))

    return data, diff


def check_cheat_equity_calc_tx(asset, golden_file, start_cash=100000000):
    """Verify equity calculations match adjusted close.

    :param asset:    Single asset to check (e.g., `'SPY'`).

    :param golden_file:   Optional path to golden file to check.

    :param start_cash=100000000:   Amount of starting cash.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    :return:  The tuple (data, diff).

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    PURPOSE:  This function puts together a DataFrame for the given `asset`
              where we immediately invest any cash deposit or dividend back
              into the asset. This is kind of cheating because:

                 1. In reality, you would not immediately receive until later.
                 2. You may not actually be able to invest all your cash
                    exactly at the current close price because you would not
                    know the close price beforehand and so would not know how
                    to setup the shares/price of a market/limit order.

              Still cheating about the above is a way to verify our
              calculations against a vendor provided adjusted close.
    """
    assert isinstance(asset, str), f'Require single asset; got {asset=}'
    data = make_full_stock_pnl_info(
        [asset], div_pay_dt_offset=0, start_cash=start_cash)

    data[name_for(asset, 'trade_flag')] = 1.0
    orig_tx_dict = make_stk_prof_tx_dict([asset], [.99])

    # Build our new tx dict starting with div cash flow
    tx_dict = {'div_cash_flow_tx': orig_tx_dict['div_cash_flow_tx']}

    # Next add a tx to spend dividends or cash deposits immediately
    # to buy as many shares (including fractional shares) as we can
    tx_dict['spend_cash_tx'] = CashToFracSharesTx(
        name_list([asset], 'close'),
        [name_for(field=n, element='overall') for n in [
            'div_cash_flow', 'cash_deposit']],
        name_list([asset], 'position'))

    # Now add transforms to calculate trade execution, trade cash flow,
    # accumulate cash, and compute equity.
    for name in ('exec_tx', 'trade_cash_flow_tx', 'cash_acc_tx', 'equity_tx'):
        tx_dict[name] = orig_tx_dict[name]

    apply_chain(data, list(tx_dict.values()))

    verify_equity_vs_adj_close(data, asset, start_cash)
    diff = check_against_golden_file(data, golden_file, (
        list(set(tx_dict['exec_tx'].input_col_names() +
                 tx_dict['exec_tx'].output_col_names()))))

    return data, diff


def verify_equity_vs_adj_close(data, asset, start_cash):
    """Verify that equity calculation matches adjusted close from start to end.

    :param data:        DataFrame with equity and adjusted close where we
                        immediately invest all cash into single asset at
                        the end of the day.

    :param asset:       Single asset we are working on.

    :param start_cash:  Initial starting cash.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    PURPOSE:  This function checks that our equity calculations match
              adjusted close. Our tolerance for the comparison depends
              on the number of dividends times times the number of
              pennies in starting cash. This is because our
              FastDivCashFlowTx cuts off fractional pennies but the
              adjusted close does not. So the difference between the
              two calculations will be larger if you use smaller
              start_cash. That error can compound over a long time but
              a few multiples of a penny per dividend is about the
              order of the acceptable error.

    """
    num_divs = len(data.loc[data[name_for(asset, 'divCash')] != 0])
    eq_name = name_for('overall', 'equity')
    equity_start = min(i for i, v in enumerate(data[eq_name].to_list()) if v)
    eq_change = data[eq_name].iloc[-1]/data[eq_name].iloc[equity_start]
    adj_close_name = name_for(asset, 'adjClose')
    adj_close_chg = data[adj_close_name].iloc[-1]/data[adj_close_name].iloc[
        equity_start]
    assert abs(adj_close_chg - eq_change) < (1.1+num_divs)/(100*start_cash)


def test_cheat_equity_spy():
    "Run check_cheat_equity_calc_tx for SPY ETF."
    check_cheat_equity_calc_tx('SPY', Path(
        tpath()) / 'data' / 'golden_cheat_SPY.csv')


def test_cheat_equity_tlt():
    "Run check_cheat_equity_calc_tx for TLT ETF."
    check_cheat_equity_calc_tx('TLT', Path(
        tpath()) / 'data' / 'golden_cheat_TLT.csv')


def test_spy_tlt():
    "Test weighting between SPY and TLT ETFs."
    check_fixme(['SPY', 'TLT'], Path(
        tpath()) / 'data' / 'golden_SPY_TLT.csv')


def test_spy_tlt_lazy():
    "Test weighting between SPY and TLT ETFs with lazy rebalance."

    asset_list = ['SPY', 'TLT']
    pre_tx_dict = {'trade_flag_tx': stk_tools.TradeFlagOnDiff(
        name_list(asset_list, 'position'),
        name_list(asset_list, 'close'),
        name_list(asset_list, 'post_trade_pos'), name_for('overall', 'equity'),
        name_list(asset_list, 'trade_flag'))}
    check_fixme(['SPY', 'TLT'], Path(
        tpath()) / 'data' / 'golden_SPY_TLT_lazy.csv', pre_tx_dict=pre_tx_dict)
    check_fixme(['SPY', 'TLT'], Path(
        tpath()) / 'data' / 'golden_SPY_TLT_lazy.csv', trade_threshold=0.01)


# pylint: disable=duplicate-code


def do_profile(preprocess_compile=False, trials=1, func_list=(
        test_cheat_equity_spy,)):  # pragma: no cover
    """Do profiling.

    :param preprocess_compile=False:  If True, we do a small run
                                      outside the profiler to force
                                      numba to JIT compile and read
                                      from the cache first.
                                      Supposedly using `cache=True` in
                                      JIT compiling will cache things,
                                      but even with that, there still
                                      seems to be an initial startup
                                      hit. See the following thread on
                                      https://numba.discourse.group/t/:

      understanding-initial-execution-delay-in-numba-cached-functions/2783/9

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    PURPOSE:  Do profiling.

    """
    if preprocess_compile:
        _ = [f() for f in func_list]

    print('Doing profiling')
    pr = cProfile.Profile()
    pr.enable()
    for _ in range(trials):
        _ = [f() for f in func_list]
    pr.disable()
    s = io.StringIO()
    sortby = SortKey.CUMULATIVE
    ps = pstats.Stats(pr, stream=s).sort_stats(sortby)
    ps.print_stats()
    print(s.getvalue())


if __name__ == '__main__':
    do_profile(preprocess_compile=True, trials=50)  # pragma: no cover



"""Module defining the FastTx class for DataFrame transformations.

# Introduction

The main goal of this module are to provide the `FastTx` class and the
`apply_chain` function to apply a chain of `FastTx` instances to a
dataframe to do some row operations.

The `FastTx` class is designed so that you can create sub-classes with
a structure that allows numba just-in-time (JIT) compilation into very
fast numerical operations. You need relatively large dataframes or
relatively complicated transforms to see the benefits of JIT, but once
you get there, the speed-ups can be pretty dramatic.

Making this work correctly is complicated because numba needs
homogenous structures (arrays and functions) to be able to JIT compile
effectively.

# Testing

If you are testing numba related code, you may want to set the
enviroment variable `NUMBA_DISABLE_JIT=1` to prevent JIT
compiling. That will allow you to use pdb and generally get better
error information.

# Motivation

The FastTx infrastructure is mainly useful when you have mixed row and
column operations. If you just have operations on rows of a column
(e.g., rolling mean), you can just call your functions/operations and
compose similar operations. You can even use numba's guvectorize to
relatively painlessly vectorize your operations.

If you have operations on columns of a row (e.g., adding one column to
another), you can just call those operations/functions as needed.

The complicated case is when you have a mix of row and column
operations and you want to be able to compose such operations or let
people mix in their own versions of such operations. As a very simple
example, imagine you have some time series data in a[t] (represented
as a pandas DataFrame) and you want to implement a system which
evolves according to:

   b[t]=c[t-1]/a[t]  c[t]=a[t]+b[t]

While c[t] is just the sum of two columns, b[t] is an operation which
depends on the previous row. So you have mixed row and column
operations. This case is trivial, but if you have a more complicated
DataFrame and want users to be able to create, compose, and calculate
mixed row and column operations, you may have poor performance in a
naive implementation.

The FastTx infrastructure is designed to let you write almost
arbitrary functions in a relatively simple way which can be easily
combined and composed but still run quickly.

While I have not experimented with Polars, I think that Polars
provides benefits mainly for cases when you can parallelize
operations. The use cases for fast_tx are mainly when you have mixed
operations like the above with many inter/intra row/column
dependancies and so parallization would not work.

# Copyright

Copyright (c) 2025, Emin Martinian - All Rights Reserved
Unauthorized copying of this file, via any medium is strictly prohibited
See AOCKS_LICENSE at the top-level of this distribution for more information
or write to emin.martinian@gmail.com for more information.

"""

import doctest
import logging
import typing

import numba
import numpy
from numpy.typing import NDArray
import pandas

from fast_tx import buffer_tools


class FastTx:
    """Fast row transform for a DataFrame.

    FastTx is an abstract class (really a protocol) for defining fast
    row transformations to apply to a dataframe. The key goals are:

      1. Enable efficient operations.
      2. Provide a structure where we can have a chain of N `FastTx`
         instances such that for each row, we can efficiently apply each
         instance to the row (including results of previous `FastTx`
         applied to the row).

    See methods for further documentation and see `apply_chain` for how
    to apply a chain of such transforms to a DataFrame.


>>> from fast_tx.core import (
...     pandas, RowShiftTx, SimpleEWA, apply_chain)
>>> df = pandas.DataFrame({
...    'a': [1.1, 2.2, 3.3, 4.4, 5.5],
...    'b': [10.0, 20.5, 30.1, 40.8, 50.2],
...    'c': [0.12, 0.34, 0.56, 0.78, 0.90],
...    'd': [1, 2, 3, 4, 5]})
>>> print(df)
     a     b     c  d
0  1.1  10.0  0.12  1
1  2.2  20.5  0.34  2
2  3.3  30.1  0.56  3
3  4.4  40.8  0.78  4
4  5.5  50.2  0.90  5
>>> apply_chain(df, [RowShiftTx(list(df)), RowShiftTx(list(df))])
>>> print(df)
      a    b    c     d
0  0.12  1.0  1.1  10.0
1  0.34  2.0  2.2  20.5
2  0.56  3.0  3.3  30.1
3  0.78  4.0  4.4  40.8
4  0.90  5.0  5.5  50.2
    """

    def __init__(self, input_col_names, output_col_names=None):
        """Initializer.

    :param input_col_names:   Either a list of strings indicating the
                              names of input columns or a dictionary
                              where keys are internal names for groups
                              of columns and values are list of strings
                              indicating the columns for that group.

                              For example, you could provide something like
                              ['price_ABC', 'price_XYZ', 'vol_ABC', 'vol_XYZ']

                              or

                              {'price': ['price_ABC', 'price_XYZ'],
                               'volume': ['vol_ABC', 'vol_XYZ']}

                              where the latter is preferable since it
                              allows using the islice method.

    :param output_col_names=None:  Optional list of output columns.

        """
        if isinstance(input_col_names, dict):
            self._input_col_dict = dict(input_col_names)
            self._input_col_names = sum(input_col_names.values(), [])
        else:
            self._input_col_dict = None
            self._input_col_names = input_col_names
        self._output_col_names = output_col_names if (
            output_col_names is not None) else input_col_names
        self._islice_dict = {}

    def startup(self, dataframe: pandas.DataFrame) -> NDArray[numpy.float64]:
        """Take dataframe as input and prepare startup. Return 1d state array.

        *IMPORTANT*:  The resulting state *MUST* be a 1-d array of float64
                      for numba JIT compiling to work correctly. You can
                      internally reshape/ravel/unravel the 1-d but when
                      passing it around it must satisfiy the type signature.

        Most transforms will not need to do anything here, but the dataframe
        is provided as input if needed. The `dataframe` should **NOT** be
        modified in any way since that is what the `calc_row` method is for.

        Mainly this should compute a state array (which must be a numpy
        array of type float64) which will be repeatedly passed to the
        `calc_row` method.
        """
        _ = dataframe
        return numpy.array([], dtype=numpy.float64)

    def shutdown(self, dataframe: pandas.DataFrame):
        """Called after we have finished processing a dataframe.

        Most transforms will not need to do anything here, but the dataframe
        is provided as input if needed. The `dataframe` should **NOT** be
        modified in any way since that is what the `calc_row` method is for.
        """

    def input_col_names(self) -> list:
        """Return a list of the input column names needed by `calc_row`.

        These input columns will be provided as the `inputs` argument
        to `calc_row`.
        """
        return self._input_col_names

    def output_col_names(self) -> list:
        """Return a list of the output column names produced by `calc_row`.

        These output columns will be provided as the `outputs` argument
        to `calc_row` which it will modify in place.
        """
        return self._output_col_names

    @staticmethod
    @numba.njit(numba.void(numba.float64, numba.float64[:],
                           numba.float64[:], numba.float64[:]), cache=True)
    def calc_row(index_value: numpy.float64, state: NDArray[numpy.float64],
                 inputs: NDArray[numpy.float64],
                 outputs: NDArray[numpy.float64]) -> None:
        """Calculate the transformed results for a row.

        :param index_value:    Value of the dataframe index we are currently
                               working on (as type numpy.float64).

        :param state:  State array produced by the `startup`
                       method. This state can be modified in place to
                       pass state to subsequent calls to `calc_row` and/or
                       `post_process_state`.

        :param inputs:  Array of inputs for the current row as named by
                        the `input_col_names` method.

        :param outputs:  Array of output for the current row as named by
                         the `output_col_names` method. The function should
                         write outputs to this array.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:   Do main calculations.

                   *IMPORTANT*:  Your function must *EXACTLY* match the type
                                 signature (including decorators) for numba
                                 JIT compiling to work.

        """
        raise ValueError('Subclass must override')  # pragma: no cover

    @staticmethod
    @numba.njit(numba.void(numba.float64, numba.float64[:],
                           numba.float64[:]), cache=True)
    def post_process_state(index_value: numpy.float64,
                           state: NDArray[numpy.float64],
                           inputs: NDArray[numpy.float64]) -> None:
        """Post process state at end of each row.

        :param index_value:    Value of the dataframe index we are currently
                               working on (as type numpy.float64).

        :param state:  State array produced by the `startup`
                       method. This state can be modified in place to
                       pass state to subsequent calls to `calc_row` and/or
                       `post_process_state`.

        :param inputs:  Array of inputs for the current row as named by
                        the `input_col_names` method.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  Sometimes you want to update your state based on the
                  final results of a row. The `calc_row` method for
                  each transform in a chain is called in order and
                  hence each transform in the chain may see different
                  inputs. After all `calc_row` methods have been called,
                  we then do a second pass and call `post_process_state`
                  for each transform. No transform should modify the
                  row via `post_process_state` (but may modify its own
                  internal state). This allows transforms to "see" the
                  final value of a row.
        """

    def islice(self, name: str):
        """Return a slice object for the column group matching `name`.

        :param name:  String name of a column group passed to `__init__`
                      (i.e., a key of the dictionary passed to `__init__`
                       for `input_col_names`).

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:  A slice object indicating the start and end of
                  that column group in `self.input_col_names()`.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  If you pass a dict to `__init__` indicating groups
                  of column names, this method gets you a slice for
                  the desired group.

        """
        result = self._islice_dict.get(name, None)
        if result is not None:
            return result
        start = 0
        for group_name, col_names in self._input_col_dict.items():
            if group_name == name:
                result = slice(start, start + len(col_names))
                self._islice_dict[name] = result
                return result
            start += len(col_names)
        raise KeyError(f'{name=} not in found in {self._input_col_dict}')

    def index(self, name: str):
        """Return index of first element in column group with given `name`.

        :param name:  String name of column group to find.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:  Starting index of first item in given column group.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  Basically like calling self.islice(name).start; see
                  docs for `islice` for details.

        """
        return self.islice(name).start

    @classmethod
    def modify_calc_row(cls):
        """Modify (or create) calc_row method.

This method will be called by `__init_subclass__`. The default version
checks if the class has a `calc_row_py` method and if so it will JIT-compile
the `calc_row_py` with numba and set `self.calc_row` to the complied
function. This makes it so you can sub-class FastTx and just define
a pure python sub-class with the `calc_row_py` method.
        """
        if hasattr(cls, 'calc_row_py'):
            logging.debug('Overriding calc_row for cls %s from calc_row_py',
                          cls)
            try:  # Try to njit with cache=True since that is faster
                cls.calc_row = staticmethod(numba.njit(
                    numba.void(numba.float64, numba.float64[:],
                               numba.float64[:], numba.float64[:]),
                    cache=True)(cls.calc_row_py))
            except RuntimeError as problem:  # but retry if cache fails
                if str(problem).startswith('cannot cache function'):
                    logging.warning(
                        'Cannot cache %s: %s; will try njit w/o cache',
                        cls.calc_row_py, problem)
                    cls.calc_row = staticmethod(numba.njit(
                        numba.void(numba.float64, numba.float64[:],
                                   numba.float64[:], numba.float64[:]),
                        cache=False)(cls.calc_row_py))
                else:
                    raise

    @classmethod
    def modify_post_process_state(cls):
        """Modify (or create) post_process_state method.

This method will be called by `__init_subclass__`. The default version
checks if the class has a `post_process_state_py` method and if so it
will JIT-compile the `post_process_state_py` with numba and set
`self.post_process_state` to the complied function. This makes it so
you can sub-class FastTx and just define a pure python sub-class with
the `post_process_state_py` method.

        """
        if hasattr(cls, 'post_process_state_py'):
            logging.debug('Overriding %s for cls %s from %s',
                          'post_process_state', cls,
                          'post_process_state_py')
            cls.post_process_state = staticmethod(numba.njit(
                numba.void(numba.float64, numba.float64[:],
                           numba.float64[:]), cache=True)(
                               cls.post_process_state_py))

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.modify_calc_row()
        cls.modify_post_process_state()


class RowShiftTx(FastTx):
    """Simple example of FastTx to just circular shift a row.


>>> from fast_tx.core import (
...     pandas, RowShiftTx, apply_chain)
>>> df = pandas.DataFrame({
...    'a': [1.1, 2.2, 3.3, 4.4, 5.5],
...    'b': [10.0, 20.5, 30.1, 40.8, 50.2],
...    'c': [0.12, 0.34, 0.56, 0.78, 0.90]})
>>> tx = RowShiftTx(['a','b','c'])
>>> apply_chain(df, [tx])
>>> print(df)
      a    b     c
0  0.12  1.1  10.0
1  0.34  2.2  20.5
2  0.56  3.3  30.1
3  0.78  4.4  40.8
4  0.90  5.5  50.2
    """

    @staticmethod
    @numba.njit(numba.void(numba.float64, numba.float64[:],
                           numba.float64[:], numba.float64[:]), cache=True)
    def calc_row(index_value: numpy.float64, state: NDArray[numpy.float64],
                 inputs: NDArray[numpy.float64],
                 outputs: NDArray[numpy.float64]) -> None:
        outputs[0] = inputs[-1]
        outputs[1:] = inputs[:-1]


class SimpleEWA(FastTx):
    """Simple example of FastTx to just circular shift a row.

    The following illustrates example usage:

>>> from fast_tx.core import (pandas, SimpleEWA, apply_chain)
>>> df = pandas.DataFrame({
...    'a': [1.1, 2.2, 3.3, 4.4, 5.5],
...    'b': [10.0, 20.5, 30.1, 40.8, 50.2],
...    'c': [0.12, 0.34, 0.56, 0.78, 0.90]})
>>> zp_df = pandas.concat([pandas.DataFrame({'a':0,'b':0,'c':0},index=[-1]),
...     df])  # zero padded for use in verifying EWM
>>> tx = SimpleEWA([.1, .2, .3], ['a','b','c'], [f'out_{n}' for n in df])
>>> apply_chain(df, [tx])
>>> print(df)
     a     b     c     out_a    out_b     out_c
0  1.1  10.0  0.12  0.110000   2.0000  0.036000
1  2.2  20.5  0.34  0.319000   5.7000  0.127200
2  3.3  30.1  0.56  0.617100  10.5800  0.257040
3  4.4  40.8  0.78  0.995390  16.6240  0.413928
4  5.5  50.2  0.90  1.445851  23.3392  0.559750
>>> all([(zp_df['a'].ewm(alpha=.1,adjust=0).mean().iloc[1:]==df.out_a).all(),
...      (zp_df['b'].ewm(alpha=.2,adjust=0).mean().iloc[1:]==df.out_b).all(),
...      (zp_df['c'].ewm(alpha=.3,adjust=0).mean().iloc[1:]==df.out_c).all()])
True
    """

    def __init__(self, alphas, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.alphas = numpy.asarray(alphas)

    def startup(self, dataframe):
        state = numpy.zeros(2*len(self._output_col_names))
        state[:len(self._output_col_names)] = self.alphas
        return state

    @staticmethod
    @numba.njit(numba.void(numba.float64, numba.float64[:],
                           numba.float64[:], numba.float64[:]), cache=True)
    def calc_row(index_value: numpy.float64, state: NDArray[numpy.float64],
                 inputs: NDArray[numpy.float64],
                 outputs: NDArray[numpy.float64]) -> None:
        # Make sure to use outputs[:] so we assign into the outputs array
        # and don't just create a new variable named `outputs`.
        alphas = state[:len(outputs)]
        prev_out = state[len(outputs):]
        outputs[:] = alphas * inputs + (1-alphas) * prev_out
        state[len(outputs):] = outputs[:]


class SafeSum(FastTx):
    """Sum inputs (ignoring na) to produce output.

    Should only use single output column since all outputs get same value.
    """

    @staticmethod
    @numba.njit(numba.void(numba.float64, numba.float64[:],
                           numba.float64[:], numba.float64[:]), cache=True)
    def calc_row(index_value: numpy.float64, state: NDArray[numpy.float64],
                 inputs: NDArray[numpy.float64],
                 outputs: NDArray[numpy.float64]) -> None:
        outputs[:] = sum([  # pylint: disable=consider-using-generator
            i for i in inputs if not numpy.isnan(i)])


class Accumulate(FastTx):
    """Sum inputs (ignoring nan) and add to previous output.

>>> from fast_tx.core import (pandas, numpy, Accumulate,
...     apply_chain)
>>> df = pandas.DataFrame({
...    'a': [1.1, 2.2, 3.3, 4.4, 5.5],
...    'b': [10.0, 20.5, 30.1, 40.8, 50.2],
...    'c': [0.12, 0.34, 0.56, 0.78, 0.90]})
>>> tx = Accumulate(['a','b','c'], ['acc'])
>>> apply_chain(df, [tx])
>>> assert df['acc'].iloc[0] == sum(df.iloc[0][['a','b','c']])
>>> assert numpy.max(numpy.abs((df.iloc[1:][['a','b','c']].sum(axis=1)
...     + df.acc.shift(1).iloc[1:]).values - df.acc.iloc[1:].values)) < 1e-9


    """

    def startup(self, dataframe):
        state = numpy.zeros(1)
        return state

    @staticmethod
    @numba.njit(numba.void(numba.float64, numba.float64[:],
                           numba.float64[:], numba.float64[:]), cache=True)
    def calc_row(index_value: numpy.float64, state: NDArray[numpy.float64],
                 inputs: NDArray[numpy.float64],
                 outputs: NDArray[numpy.float64]) -> None:
        outputs[0] = sum([  # pylint: disable=consider-using-generator
            i for i in inputs if not numpy.isnan(i)]) + state[0]
        state[0] = outputs[0]


class Lag(FastTx):
    """Simple example of FastTx to lag a row (like shift in pandas).

    This is slightly tricky because what we want to do is capture the
    value of a row *AFTER* all calculations and then apply it *BEFORE*
    new calculations. This requires implementing post_process_state.

    The following illustrates example usage:

>>> from fast_tx.core import (pandas, Lag, SafeSum,
...     apply_chain)
>>> df = pandas.DataFrame({
...    'a': [1.1, 2.2, 3.3, 4.4, 5.5],
...    'b': [10.0, 20.5, 30.1, 40.8, 50.2],
...    'c': [0.12, 0.34, 0.56, 0.78, 0.90]})
>>> tx = Lag(1, ['a','b','d'], ['lag_a', 'lag_b', 'lag_d'])
>>> apply_chain(df, [tx, SafeSum(['lag_b','c'], ['d']),
...     SafeSum(['lag_d','c'], ['e'])])
>>> print(df)
     a     b     c  lag_a  lag_b  lag_d      d      e
0  1.1  10.0  0.12    0.0    0.0   0.00   0.12   0.12
1  2.2  20.5  0.34    1.1   10.0   0.12  10.34   0.46
2  3.3  30.1  0.56    2.2   20.5  10.34  21.06  10.90
3  4.4  40.8  0.78    3.3   30.1  21.06  30.88  21.84
4  5.5  50.2  0.90    4.4   40.8  30.88  41.70  31.78

Note in the above that `lag_d` is correct. This may seem trivial, but
is a bit tricky (and hence why we needed to implement
post_process_state) since `lag_d` is computed by the 2nd transform
(which comes **AFTER** the Lag transform). The Lag transform must come
**BEFORE** everything else in the chain so it can put its values into
the row before the other transforms process the row, but the Lag
transform must update its state **AFTER** all calculations so it
captures the final state of the row. This is achieved by having a
seprate function for post_process_state vs calc_row.


We can also verify that Lag matches pandas shift:


>>> for s in range(1, 4):  # check lag matches shift
...     apply_chain(df, [Lag(s, tx.input_col_names(), tx.output_col_names())])
...     assert not (df.shift(s)[tx.input_col_names()].values
...                 - df[tx.output_col_names()].values)[s:].any()
...

    """
    def __init__(self, lag_length, *args, **kwargs):
        self.lag_length = lag_length
        assert lag_length > 0
        super().__init__(*args, **kwargs)

    def startup(self, dataframe):
        state = numpy.zeros(1 + self.lag_length * len(self._input_col_names))
        state[-1] = self.lag_length
        return state

    @staticmethod
    @numba.njit(numba.void(numba.float64, numba.float64[:],
                           numba.float64[:], numba.float64[:]), cache=True)
    def calc_row(index_value: numpy.float64, state: NDArray[numpy.float64],
                 inputs: NDArray[numpy.float64],
                 outputs: NDArray[numpy.float64]) -> None:
        outputs[:] = state[-len(outputs)-1:-1]

    @staticmethod
    @numba.njit(numba.void(numba.float64, numba.float64[:],
                           numba.float64[:]), cache=True)
    def post_process_state(index_value, state, inputs):
        lag_length = state[-1]
        if lag_length > 1:
            state[len(inputs):-1] = state[
                :-len(inputs)-1]
        state[:len(inputs)] = inputs[:]


def apply_chain(dataframe: pandas.DataFrame,
                tx_chain: typing.List[FastTx]):
    """Apply given transform chain to dataframe.

    :param dataframe:  DataFrame to modify.

    :param tx_chain:  List of `FastTx` instances to apply to dataframe.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    PURPOSE:  Apply the `tx_chain`.  See docs for `FastTx` and example
              transforms like `Lag` for example usage.

    """
    input_cols, output_cols, maincols = buffer_tools.prepare_cols(
        dataframe, tx_chain)
    states, state_len = buffer_tools.prep_states_array(tx_chain, dataframe)

    index, input_buffer, output_buffer, mainbuffer = buffer_tools.make_buffers(
        dataframe, maincols, input_cols, output_cols)

    calc_tx_list, post_process_tx_list = buffer_tools.prepare_tx_lists(
        tx_chain, FastTx.post_process_state,
        buffer_tools.noop_post_process_state)

    # If original dataframe has mixed types, mainbuffer may not be float.
    # We must force float for numba, but warn the user about that so they
    # can maybe try to fix type themselves.

    if mainbuffer.dtype != 'float64':  # pragma: no cover
        logging.warning('Buffer dataframe has type %s; will force float64',
                        mainbuffer.dtype)
        mainbuffer = mainbuffer.astype(numpy.float64)

    buffer_tools.apply_inner_loop(
        index.astype(numpy.float64), mainbuffer, input_buffer, output_buffer,
        states, state_len, calc_tx_list, post_process_tx_list,
        input_cols, output_cols)
    dataframe[list(maincols)] = mainbuffer

    for tx in tx_chain:
        tx.shutdown(dataframe)


if __name__ == '__main__':    # pragma: no cover
    doctest.testmod()         # pragma: no cover
    print('Finished tests')   # pragma: no cover
